# -*- coding: utf-8 -*-
#
"""
Most of the XSD datatypes are handled directly by RDFLib. However, in some cases, that is not good enough. There are two
major reasons for this:

#. Some datatypes are missing from RDFLib and required by OWL 2 RL and/or RDFS.
#. In other cases, though the datatype is present, RDFLib is fairly lax in checking the lexical value of those datatypes. Typical case is boolean.

Some of these deficiencies are handled by this module. All the functions convert the lexical value into a
python datatype (or return the original string if this is not possible) which will be used, e.g.,
for comparisons (equalities). If the lexical value constraints are not met, exceptions are raised.

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/
"""

__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

# noinspection PyPep8Naming
# from owlrl.RDFS import RDFNS as ns_rdf
from rdflib.namespace import RDF, XSD

from rdflib.term import XSDToPython, Literal, _toPythonMapping

import datetime, time, re
from decimal import Decimal


# noinspection PyMissingConstructor,PyPep8Naming
class _namelessTZ(datetime.tzinfo):
    """
    (Nameless) timezone object. The python datetime object requires timezones as
    a specific object added to the conversion, rather than the explicit hour and minute
    difference used by XSD. This class is used to wrap around the hour/minute values.

    :param hours: Hour offset.
    :param minutes: Minute offset
    """

    def __init__(self, hours, minutes):
        """
        @param hours: hour offset
        @param minutes: minute offset
        """
        self.__offset = datetime.timedelta(hours=hours, minutes=minutes)
        self.__name = "nameless"

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return self.__name

    def dst(self, dt):
        return datetime.timedelta(0)


# noinspection PyPep8Naming
def _returnTimeZone(incoming_v):
    """Almost all time/date related methods require the extraction of optional time zone information.
    @param incoming_v: the time/date string
    @return (v,timezone) tuple; 'v' is the input string with the timezone info cut off, 'timezone' is a L{_namelessTZ}
    instance or None
    """
    if incoming_v[-1] == "Z":
        v = incoming_v[:-1]
        tzone = _namelessTZ(0, 0)
    else:
        pattern = r".*(\+|-)([0-9][0-9]):([0-9][0-9])"
        match = re.match(pattern, incoming_v)
        if match is None:
            v = incoming_v
            tzone = None
        else:
            hours = int(match.groups()[1])
            if match.groups()[0] == "-":
                hours = -hours - 1
            minutes = int(match.groups()[2])
            v = incoming_v[:-6]
            tzone = _namelessTZ(hours, minutes)
    return v, tzone


# Booleans ##################################################
# noinspection PyPep8Naming
def _strToBool(v):
    """The built-in conversion to boolean is way too lax. The xsd specification requires that only true, false, 1 or 0 should be used...
    @param v: the literal string defined as boolean
    @return corresponding boolean value
    @raise ValueError: invalid boolean values
    """
    if v.lower() == "true" or v.lower() == "1":
        return True
    elif v.lower() == "false" or v.lower() == "0":
        return False
    else:
        raise ValueError("Invalid boolean literal value %s" % v)


# Decimals ##################################################
# noinspection PyPep8Naming
def _strToDecimal(v):
    """The built in datatype handling for RDFLib maps a decimal number to float, but the python version 2.4 and upwards
    also has a Decimal number. Better make use of that to use very high numbers.
    However, there is also a big difference between Python's decimal and XSD's decimal, because the latter does not
    allow for an exponential normal form (why???). This must be filtered out.
    @param v: the literal string defined as decimal
    @return Decimal
    @raise ValueError: invalid decimal value
    """
    # check whether the lexical form of 'v' is o.k.
    if v.find("E") != -1 or v.find("e") != -1:
        # this is an invalid lexical form, though would be accepted by Python
        raise ValueError("Invalid decimal literal value %s" % v)
    else:
        return Decimal(v)


# ANY URIS ##################################################
# set of characters allowed in a hexadecimal number
_hexc = ["A", "B", "C", "D", "E", "F", "a", "b", "c", "d", "e", "f"]
# set of numerals
_numb = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "0"]
# noinspection PyPep8Naming
def _strToAnyURI(v):
    """Rudimentary test for the AnyURI value. If it is a relative URI, then some tests are done to filter out
    mistakes. I am not sure this is the full implementation of the RFC, though, may have to be checked at some point
    later.
    @param v: the literal string defined as a URI
    @return the incoming value
    @raise ValueError: invalid URI value
    """
    import urllib.parse

    if len(v) == 0:
        return v
    if urllib.parse.urlsplit(v)[0] != "":
        # this means that there is a proper scheme, the URI should be kosher
        return v
    else:
        # this is meant to be a relative URI.
        # If I am correct, that cannot begin with one or more "?" or ":" characters
        # all others are o.k.
        # if it begins with a % then it should be followed by two hexa characters,
        # otherwise it is also a bug
        if v[0] == "%":
            if (
                len(v) >= 3
                and (v[1] in _hexc or v[1] in _numb)
                and (v[2] in _hexc or v[2] in _numb)
            ):
                return v
            else:
                raise ValueError("Invalid IRI %s" % v)
        elif v[0] == "?" or v[0] == ":":
            raise ValueError("Invalid IRI %s" % v)
        else:
            return v


# Base64Binary ##################################################
# noinspection PyPep8Naming
def _strToBase64Binary(v):
    """Rudimentary test for the base64Binary value. The problem is that the built-in b64 module functions ignore the
    fact that only a certain family of characters are allowed to appear in the lexical value, so this is checked first.
    @param v: the literal string defined as a base64encoded string
    @return the decoded (binary) content
    @raise ValueError: invalid base 64 binary value
    """
    import base64

    if v.replace("=", "x").replace("+", "y").replace("/", "z").isalnum():
        try:
            return base64.standard_b64decode(v)
        except:
            raise ValueError("Invalid Base64Binary %s" % v)
    else:
        raise ValueError("Invalid Base64Binary %s" % v)


# Numerical types ##################################################
# limits for unsigned bytes
_limits_unsignedByte = [-1, 256]


# limits for bytes
_limits_byte = [-129, 128]


# limits for unsigned int
_limits_unsignedInt = [-1, 4294967296]


# limits for int
_limits_int = [-2147483649, 2147483648]


# limits for unsigned short
_limits_unsignedShort = [-1, 65536]


# limits for short
_limits_short = [-32769, 32768]


# limits for unsigned long
_limits_unsignedLong = [-1, 18446744073709551616]


# limits for long
_limits_long = [-9223372036854775809, 9223372036854775808]


# limits for positive integer
_limits_positiveInteger = [0, None]


# limits for non positive integer
_limits_nonPositiveInteger = [None, 1]


# limits for non negative integer
_limits_nonNegativeInteger = [-1, None]


# limits for negative integer
_limits_negativeInteger = [None, 0]


# noinspection PyPep8Naming,PyBroadException
def _strToBoundNumeral(v, interval, conversion):
    """Test (and convert) a generic numerical type, with a check against a lower and upper limit.
    @param v: the literal string to be converted
    @param interval: lower and upper bounds (non inclusive). If the value is None, no comparison should be done
    @param conversion: conversion function, ie, int, long, etc
    @raise ValueError: invalid value
    """
    try:
        i = conversion(v)
        if (interval[0] is None or interval[0] < i) and (
            interval[1] is None or i < interval[1]
        ):
            return i
    except:
        pass
    raise ValueError("Invalid numerical value %s" % v)


# Double and float ##################################################
# noinspection PyPep8Naming
def _strToDouble(v):
    """Test and convert a double value into a Decimal or float. Raises an exception if the number is outside the
    permitted range, ie, 1.0E+310 and 1.0E-330. To be on the safe side (python does not have double!) Decimals are used
    if possible. Upper and lower values, as required by xsd, are checked (and these fixed values are the reasons
    why Decimal is used!)

    @param v: the literal string defined as a double
    @return Decimal
    @raise ValueError: invalid value
    """
    try:
        value = Decimal(v)
        upper = Decimal("1.0E+310")
        lower = Decimal("1.0E-330")
        if lower < abs(value) < upper:
            # bingo
            return value
        else:
            raise ValueError("Invalid double %s" % v)
    except:
        # there was a problem in creating a decimal...
        raise ValueError("Invalid double %s" % v)


# noinspection PyPep8Naming
def _strToFloat(v):
    """Test and convert a float value into Decimal or (python) float. Raises an exception if the number is outside the
    permitted range, ie, 1.0E+40 and 1.0E-50. (And these fixed values are the reasons why Decimal is used!)

    @param v: the literal string defined as a float
    @return Decimal if the local python version is >= 2.4, float otherwise
    @raise ValueError: invalid value
    """
    try:
        value = Decimal(v)
        upper = Decimal("1.0E+40")
        lower = Decimal("1.0E-50")
        if lower < abs(value) < upper:
            # bingo
            return value
        else:
            raise ValueError("Invalid float %s" % v)
    except:
        # there was a problem in creating a decimal...
        raise ValueError("Invalid float %s" % v)


# hexa ##################################################
# noinspection PyPep8Naming
def _strToHexBinary(v):
    """Test (and convert) hexa integer values. The number of characters should be even.
    @param v: the literal string defined as a hexa number
    @return long value
    @raise ValueError: invalid value
    """
    # first of all, the number of characters must be even according to the xsd spec:
    length = len(v)
    if (length / 2) * 2 != length:
        raise ValueError("Invalid hex binary number %s" % v)
    return int(v, 16)


# Datetime, date timestamp, etc ################################
# noinspection PyPep8Naming
def _strToDateTimeAndStamp(incoming_v, timezone_required=False):
    """Test (and convert) datetime and date timestamp values.
    @param incoming_v: the literal string defined as the date and time
    @param timezone_required: whether the timezone is required (ie, for date timestamp) or not
    @return datetime
    @rtype: datetime.datetime
    @raise ValueError: invalid datetime or date timestamp
    """

    # First, handle the timezone portion, if there is any
    (v, tzone) = _returnTimeZone(incoming_v)

    # Check on the timezone. For time date stamp object it is required
    if timezone_required and tzone is None:
        raise ValueError("Invalid datetime %s" % incoming_v)

    # The microseconds should be handled here...
    final_v = v
    milliseconds = 0
    milpattern = r"(.*)(\.)([0-9]*)"
    match = re.match(milpattern, v)
    if match is not None:
        # we have a millisecond portion...
        try:
            final_v = match.groups()[0]
            milliseconds = int(match.groups()[2])
        except:
            raise ValueError("Invalid datetime %s" % incoming_v)
    #
    # By now, the pattern should be clear
    # This may raise an exception...
    try:
        tstr = time.strptime(final_v, "%Y-%m-%dT%H:%M:%S")
        if tzone is not None:
            return datetime.datetime(
                tstr.tm_year,
                tstr.tm_mon,
                tstr.tm_mday,
                tstr.tm_hour,
                tstr.tm_min,
                tstr.tm_sec,
                milliseconds,
                tzone,
            )
        else:
            return datetime.datetime(
                tstr.tm_year,
                tstr.tm_mon,
                tstr.tm_mday,
                tstr.tm_hour,
                tstr.tm_min,
                tstr.tm_sec,
                milliseconds,
            )
    except:
        raise ValueError("Invalid datetime %s" % incoming_v)


# noinspection PyPep8Naming
def _strToTime(incoming_v):
    """Test (and convert) time values.
    @param incoming_v: the literal string defined as time value
    @return time
    @rtype datetime.time
    @raise ValueError: invalid datetime or date timestamp
    """

    # First, handle the timezone portion, if there is any
    (v, tzone) = _returnTimeZone(incoming_v)

    # The microseconds should be handled here...
    final_v = v
    milliseconds = 0
    milpattern = r"(.*)(\.)([0-9]*)"
    match = re.match(milpattern, v)
    if match is not None:
        # we have a millisecond portion...
        try:
            final_v = match.groups()[0]
            milliseconds = int(match.groups()[2])
        except:
            raise ValueError("Invalid datetime %s" % incoming_v)
    #
    # By now, the pattern should be clear
    # This may raise an exception...
    try:
        tstr = time.strptime(final_v, "%H:%M:%S")
        if tzone is not None:
            return datetime.time(
                tstr.tm_hour, tstr.tm_min, tstr.tm_sec, milliseconds, tzone
            )
        else:
            return datetime.time(tstr.tm_hour, tstr.tm_min, tstr.tm_sec, milliseconds)
    except:
        raise ValueError("Invalid time %s" % incoming_v)


# noinspection PyPep8Naming
def _strToDate(incoming_v):
    """Test (and convert) date values.
    @param incoming_v: the literal string defined as date (in iso format)
    @return date
    @return datetime.date
    @raise ValueError: invalid datetime or date timestamp
    """

    # First, handle the timezone portion, if there is any
    (final_v, tzone) = _returnTimeZone(incoming_v)

    # This may raise an exception...
    try:
        tstr = time.strptime(final_v, "%Y-%m-%d")
        return datetime.date(tstr.tm_year, tstr.tm_mon, tstr.tm_mday)
    except:
        raise ValueError("Invalid date %s" % incoming_v)


# The 'g' series for dates ############################
# The 'g' datatypes (eg, gYear) cannot be directly represented as a python datatype
# the series of methods below simply check whether the incoming string is o.k., but the
# returned value is the same as the original
# noinspection PyPep8Naming
def _strTogYearMonth(v):
    """Test gYearMonth value
    @param v: the literal string
    @return v
    @raise ValueError: invalid value
    """
    try:
        time.strptime(v + "-01", "%Y-%m-%d")
        return v
    except:
        raise ValueError("Invalid gYearMonth %s" % v)


# noinspection PyPep8Naming
def _strTogYear(v):
    """Test gYear value
    @param v: the literal string
    @return v
    @raise ValueError: invalid value
    """
    try:
        time.strptime(v + "-01-01", "%Y-%m-%d")
        return v
    except:
        raise ValueError("Invalid gYear %s" % v)


# noinspection PyPep8Naming
def _strTogMonthDay(v):
    """Test gYearMonth value
    @param v: the literal string
    @return v
    @raise ValueError: invalid value
    """
    try:
        time.strptime("2008-" + v, "%Y-%m-%d")
        return v
    except:
        raise ValueError("Invalid gMonthDay %s" % v)


# noinspection PyPep8Naming
def _strTogDay(v):
    """Test gYearMonth value
    @param v: the literal string
    @return v
    @raise ValueError: invalid value
    """
    try:
        time.strptime("2001-01-" + v, "%Y-%m-%d")
        return v
    except:
        raise ValueError("Invalid gDay %s" % v)


# noinspection PyPep8Naming
def _strTogMonth(v):
    """Test gYearMonth value
    @param v: the literal string
    @return v
    @raise ValueError: invalid value
    """
    try:
        time.strptime("2001-" + v + "-01", "%Y-%m-%d")
        return v
    except:
        raise ValueError("Invalid gMonth %s" % v)


# XML Literal #########################################
# noinspection PyPep8Naming
def _strToXMLLiteral(v):
    """Test (and convert) XML Literal values.
    @param v: the literal string defined as an xml literal
    @return the canonical version of the same xml text
    @raise ValueError: incorrect xml string
    """
    import xml.dom.minidom

    try:
        dom = xml.dom.minidom.parseString(v)
        return dom.toxml()
    except:
        raise ValueError("Invalid XML Literal %s" % v)


# language, NMTOKEN, NAME, etc #########################
# regular expression for a 'language' datatype
_re_language = r"[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*"


# regexp for NMTOKEN. It must be used with a re.U flag (the '(?U' regexp form did not work. It may depend on the
# locale...)
_re_NMTOKEN = r"[\w:_.\-]+"


# characters not permitted at a starting position for Name (otherwise Name is like NMTOKEN
_re_Name_ex = [".", "-"] + _numb


# regexp for NCName. It must be used with a re.U flag (the '(?U' regexp form did not work. It may depend on the
# locale...)
_re_NCName = r"[\w_.\-]+"


# characters not permitted at a starting position for NCName
_re_NCName_ex = [".", "-"] + _numb


# noinspection PyDefaultArgument,PyPep8Naming,PyPep8Naming
def _strToVal_Regexp(v, regexp, flag=0, excludeStart=[]):
    """Test (and convert) a generic string type, with a check against a regular expression.
    @param v: the literal string to be converted
    @param regexp: the regular expression to check against
    @param flag: flags to be used in the regular expression
    @param excludeStart: array of characters disallowed in the first position
    @return original string
    @raise ValueError: invalid value
    """
    match = re.match(regexp, v, flag)
    if match is None or match.end() != len(v):
        raise ValueError("Invalid literal %s" % v)
    else:
        if len(excludeStart) > 0 and v[0] in excludeStart:
            raise ValueError("Invalid literal %s" % v)
        return v


# Disallowed characters in a token or a normalized string, as a regexp
_re_token = "[^\n\t\r]+"


# noinspection PyPep8Naming
def _strToToken(v):
    """Test (and convert) a string to a token.
    @param v: the literal string to be converted
    @return original string
    @raise ValueError: invalid value
    """
    if len(v) == 0:
        return v
    # filter out the case when there are new lines and similar (if there is a problem, an exception is raised)
    _strToVal_Regexp(v, _re_token)
    v1 = " ".join(v.strip().split())
    # normalize the string, and see if the result is the same:
    if len(v1) == len(v):
        # no characters lost, ie, no unnecessary spaces
        return v
    else:
        raise ValueError("Invalid literal %s" % v)


# plain literal ########################################
# noinspection PyPep8Naming
def _strToPlainLiteral(v):
    """Test (and convert) a plain literal
    @param v: the literal to be converted
    @return a new RDFLib Literal with language tag
    @raise ValueError: invalid value
    """
    reg = "(.*)@([^@]*)"
    # a plain literal must match this regexp!
    match = re.match(reg, v)
    if match is None:
        raise ValueError("Invalid plain literal %s" % v)
    else:
        lit = match.groups()[0]
        if len(match.groups()) == 1 or match.groups()[1] == "":
            # no language tag
            return Literal(lit)
        else:
            lang = match.groups()[1]
            # check if this is a correct language tag. Note that can raise an exception!
            try:
                lang = _strToVal_Regexp(lang, _re_language)
                return Literal(lit, lang=lang.lower())
            except:
                raise ValueError("Invalid plain literal %s" % v)


#####################################################################################
# Replacement of RDFLib's conversion function. Each entry assigns a function to an XSD datatype, attempting to convert
# a string to a Python datatype (or raise an exception if some problem is found)
AltXSDToPYTHON = {
    XSD.language: lambda v: _strToVal_Regexp(v, _re_language),
    XSD.NMTOKEN: lambda v: _strToVal_Regexp(v, _re_NMTOKEN, re.U),
    XSD.Name: lambda v: _strToVal_Regexp(v, _re_NMTOKEN, re.U, _re_Name_ex),
    XSD.NCName: lambda v: _strToVal_Regexp(v, _re_NCName, re.U, _re_NCName_ex),
    XSD.token: _strToToken,
    RDF.PlainLiteral: _strToPlainLiteral,
    XSD.boolean: _strToBool,
    XSD.decimal: _strToDecimal,
    XSD.anyURI: _strToAnyURI,
    XSD.base64Binary: _strToBase64Binary,
    XSD.double: _strToDouble,
    XSD.float: _strToFloat,
    XSD.byte: lambda v: _strToBoundNumeral(v, _limits_byte, int),
    XSD.int: lambda v: _strToBoundNumeral(v, _limits_int, int),
    XSD.long: lambda v: _strToBoundNumeral(v, _limits_long, int),
    XSD.positiveInteger: lambda v: _strToBoundNumeral(v, _limits_positiveInteger, int),
    XSD.nonPositiveInteger: lambda v: _strToBoundNumeral(
        v, _limits_nonPositiveInteger, int
    ),
    XSD.negativeInteger: lambda v: _strToBoundNumeral(v, _limits_negativeInteger, int),
    XSD.nonNegativeInteger: lambda v: _strToBoundNumeral(
        v, _limits_nonNegativeInteger, int
    ),
    XSD.short: lambda v: _strToBoundNumeral(v, _limits_short, int),
    XSD.unsignedByte: lambda v: _strToBoundNumeral(v, _limits_unsignedByte, int),
    XSD.unsignedShort: lambda v: _strToBoundNumeral(v, _limits_unsignedShort, int),
    XSD.unsignedInt: lambda v: _strToBoundNumeral(v, _limits_unsignedInt, int),
    XSD.unsignedLong: lambda v: _strToBoundNumeral(v, _limits_unsignedLong, int),
    XSD.hexBinary: _strToHexBinary,
    XSD.dateTime: lambda v: _strToDateTimeAndStamp(v, False),
    XSD.dateTimeStamp: lambda v: _strToDateTimeAndStamp(v, True),
    RDF.XMLLiteral: _strToXMLLiteral,
    XSD.integer: int,
    XSD.string: lambda v: v,
    RDF.HTML: lambda v: v,
    XSD.normalizedString: lambda v: _strToVal_Regexp(v, _re_token),
    # These are RDFS specific...
    XSD.time: _strToTime,
    XSD.date: _strToDate,
    XSD.gYearMonth: _strTogYearMonth,
    XSD.gYear: _strTogYear,
    XSD.gMonthDay: _strTogMonthDay,
    XSD.gDay: _strTogDay,
    XSD.gMonth: _strTogMonth,
}


def use_Alt_lexical_conversions():
    """
    Registering the datatypes item for RDFLib, ie, bind the dictionary values. The 'bind' method of RDFLib adds
    extra datatypes to the registered ones in RDFLib, though the table used here (I.e., :code:`AltXSDToPYTHON`) actually
    overrides all of the default conversion routines. The method also add a Decimal entry to the :code:`PythonToXSD` list of
    RDFLib.
    """
    _toPythonMapping.update(AltXSDToPYTHON)


def use_RDFLib_lexical_conversions():
    """
    Restore the original (ie, RDFLib) set of lexical conversion routines.
    """
    _toPythonMapping.update(XSDToPython)


#######################################################################################
# This module can pretty much tested individually...


if __name__ == "__main__":
    import sys

    dtype = sys.argv[1]
    string = sys.argv[2]
    datatype = XSD[dtype]
    result = AltXSDToPYTHON[datatype](string)
    print(type(result))
    print(result)



# -*- coding: utf-8 -*-
#
"""
Module to datatype restrictions, i.e., data ranges.
 
The module implements the following aspects of datatype restrictions:

    - a new datatype is created run-time and added to the allowed and accepted datatypes; literals are checked whether they abide to the restrictions
    - the new datatype is defined to be a 'subClass' of the restricted datatype
    - literals of the restricted datatype and that abide to the restrictions defined by the facets are also assigned to be of the new type
 
The last item is important to handle the following structures::

 ex:RE a owl:Restriction ;
    owl:onProperty ex:p ;
    owl:someValuesFrom [
        a rdfs:Datatype ;
        owl:onDatatype xsd:string ;
        owl:withRestrictions (
            [ xsd:minLength "3"^^xsd:integer ]
            [ xsd:maxLength "6"^^xsd:integer ]
        )
    ]
 .
 ex:q ex:p "abcd"^^xsd:string.

In the case above the system can then infer that :code:`ex:q` is also of type :code:`ex:RE`.

Datatype restrictions are used by the :class:`.OWLRLExtras.OWLRL_Extension` extension class.

The implementation is **not** 100% complete. Some things that an ideal implementation should do are not done yet like:

    - checking whether a facet is of a datatype that is allowed for that facet
    - handling of non-literals in the facets (ie, if the resource is defined to be of type literal, but whose value is defined via a separate :code:`owl:sameAs` somewhere else)

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""
__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

import re

from rdflib.namespace import OWL, RDF, RDFS, XSD
from rdflib import Literal as rdflibLiteral, Graph

from .DatatypeHandling import AltXSDToPYTHON
from functools import reduce

# Constant for datatypes using min, max (inclusive and exclusive):
MIN_MAX = 0
# Constant for datatypes using length, minLength, and maxLength (and nothing else)
LENGTH = 1
# Constant for datatypes using length, minLength, maxLength, and pattern
LENGTH_AND_PATTERN = 2
# Constant for datatypes using length, minLength, maxLength, pattern, and lang range
LENGTH_PATTERN_LRANGE = 3

# Dictionary of all the datatypes, keyed by category
Datatypes_per_facets = {
    MIN_MAX: [
        # OWL.rational,
        # OWL.real,
        XSD.decimal,
        XSD.integer,
        XSD.nonNegativeInteger,
        XSD.nonPositiveInteger,
        XSD.positiveInteger,
        XSD.negativeInteger,
        XSD.long,
        XSD.short,
        XSD.byte,
        XSD.unsignedLong,
        XSD.unsignedInt,
        XSD.unsignedShort,
        XSD.unsignedByte,
        XSD.double,
        XSD.float,
        XSD.dateTime,
        XSD.dateTimeStamp,
        XSD.time,
        XSD.date,
    ],
    LENGTH: [XSD.hexBinary, XSD.base64Binary],
    LENGTH_AND_PATTERN: [
        XSD.anyURI,
        XSD.string,
        XSD.NMTOKEN,
        XSD.Name,
        XSD.NCName,
        XSD.language,
        XSD.normalizedString,
    ],
    LENGTH_PATTERN_LRANGE: [RDF.PlainLiteral],
}

# a simple list containing C{all} datatypes that may have a facet
facetable_datatypes = reduce(lambda x, y: x + y, list(Datatypes_per_facets.values()))

#######################################################################################################


def _lit_to_value(dt, v):
    """
    This method is used to convert a string to a value with facet checking. RDF Literals are converted to
    Python values using this method; if there is a problem, an exception is raised (and caught higher
    up to generate an error message).

    The method is the equivalent of all the methods in the :mod:`.DatatypeHandling` module, and is registered
    to the system run time, as new restricted datatypes are discovered.

    (Technically, the registration is done via a :code:`lambda v: _lit_to_value(self,v)` setting from within a
    :class:`.RestrictedDatatype` instance).

    :param dt: Faceted datatype.
    :type dt: :class:`RestrictedDatatype`

    :param v: Literal to be converted and checked.

    :raise ValueError: Invalid literal value.
    """
    # This may raise an exception...
    value = dt.converter(v)

    # look at the different facet categories and try to find which is
    # is, if any, the one that is of relevant for this literal
    for cat in Datatypes_per_facets:
        if dt.base_type in Datatypes_per_facets[cat]:
            # yep, this is to be checked.
            if not dt.checkValue(value):
                raise ValueError(
                    "Literal value %s does not fit the faceted datatype %s" % (v, dt)
                )
    # got here, everything should be fine
    return value


# noinspection PyPep8Naming,PyShadowingBuiltins
def _lang_range_check(range, lang):
    """
    Implementation of the extended filtering algorithm, as defined in point 3.3.2,
    of U{RFC 4647<http://www.rfc-editor.org/rfc/rfc4647.txt>}, on matching language ranges and language tags.
    Needed to handle the C{rdf:PlainLiteral} datatype.
    @param range: language range
    @param lang: language tag
    @rtype: boolean
    """

    def _match(r, l):
        """Matching of a range and language item: either range is a wildcard or the two are equal
        @param r: language range item
        @param l: language tag item
        @rtype: boolean
        """
        return r == "*" or r == l

    rangeList = range.strip().lower().split("-")
    langList = lang.strip().lower().split("-")
    if not _match(rangeList[0], langList[0]):
        return False

    rI = 1
    rL = 1
    while rI < len(rangeList):
        if rangeList[rI] == "*":
            rI += 1
            continue
        if rL >= len(langList):
            return False
        if _match(rangeList[rI], langList[rL]):
            rI += 1
            rL += 1
            continue
        if len(langList[rL]) == 1:
            return False
        else:
            rL += 1
            continue
    return True


#######################################################################################################


def extract_faceted_datatypes(core, graph: Graph):
    """
    Extractions of restricted (i.e., faceted) datatypes from the graph.

    :param core: The core closure instance that is being handled.
    :type core: :class:`.Closure.Core`

    :param graph: RDFLib graph.
    :type graph: :class:`rdflib.graph.Graph`

    :return: List of :class:`.RestrictedDatatype` instances.
    :rtype: list
    """
    retval = []
    for dtype in graph.subjects(RDF.type, RDFS.Datatype):
        base_type = None
        facets = []
        try:
            base_types = [x for x in graph.objects(dtype, OWL.onDatatype)]
            if len(base_types) > 0:
                if len(base_types) > 1:
                    raise Exception(
                        "Several base datatype for the same restriction %s" % dtype
                    )
                else:
                    base_type = base_types[0]
                    if base_type in facetable_datatypes:
                        rlists = [x for x in graph.objects(dtype, OWL.withRestrictions)]
                        if len(rlists) > 1:
                            raise Exception(
                                "More than one facet lists for the same restriction %s"
                                % dtype
                            )
                        elif len(rlists) > 0:
                            final_facets = []
                            for r in graph.items(rlists[0]):
                                for (facet, lit) in graph.predicate_objects(r):
                                    if isinstance(lit, rdflibLiteral):
                                        # the python value of the literal should be extracted
                                        # note that this call may lead to an exception, but that is fine,
                                        # it is caught some lines below anyway...
                                        try:
                                            if (
                                                lit.datatype is None
                                                or lit.datatype == XSD.string
                                            ):
                                                final_facets.append((facet, str(lit)))
                                            else:
                                                final_facets.append(
                                                    (
                                                        facet,
                                                        AltXSDToPYTHON[lit.datatype](
                                                            str(lit)
                                                        ),
                                                    )
                                                )
                                        except Exception as msg:
                                            core.add_error(msg)
                                            continue
                                # We do have everything we need:
                            new_datatype = RestrictedDatatype(
                                dtype, base_type, final_facets
                            )
                            retval.append(new_datatype)
        except Exception as msg:
            # import sys
            # print sys.exc_info()
            # print sys.exc_type
            # print sys.exc_value
            # print sys.exc_traceback
            core.add_error(msg)
            continue
    return retval


# noinspection PyPep8Naming
class RestrictedDatatypeCore:
    """
    An 'abstract' superclass for datatype restrictions. The instance variables listed here are
    used in general, without the specificities of the concrete restricted datatype.

    This module defines the :class:`.RestrictedDatatype` class that corresponds to the datatypes and their restrictions
    defined in the OWL 2 standard. Other modules may subclass this class to define new datatypes with restrictions.

    :ivar type_uri: The URI for this datatype.

    :ivar base_type: URI of the datatype that is restricted.

    :ivar toPython: Function to convert a Literal of the specified type to a Python value.
    """

    def __init__(self, type_uri, base_type):
        self.datatype = type_uri
        self.base_type = base_type
        self.toPython = None

    def checkValue(self, value):
        """
        Check whether the (Python) value abides to the constraints defined by the current facets.

        :param value: The value to be checked.
        :rtype: bool
        """
        raise Exception(
            "This class should not be used by itself, only via its subclasses!"
        )


# noinspection PyPep8Naming
class RestrictedDatatype(RestrictedDatatypeCore):
    """
    Implementation of a datatype with facets, ie, datatype with restrictions.

    :param type_uri: URI of the datatype being defined.
    :param base_type: URI of the base datatype, ie, the one being restricted.
    :param facets: List of :code:`(facetURI, value)` pairs.

    :ivar datatype : The URI for this datatype.

    :ivar base_type: URI of the datatype that is restricted.

    :ivar converter: Method to convert a literal of the base type to a Python value (:code:`DatatypeHandling.AltXSDToPYTHON`).

    :ivar minExclusive: Value for the :code`xsd:minExclusive` facet, initialized to :code:`None` and set to the right value if
        a facet is around.
    :ivar minInclusive: Value for the :code:`xsd:minInclusive` facet, initialized to :code:`None` and set to the right value if
        a facet is around.
    :ivar maxExclusive: Value for the :code:`xsd:maxExclusive` facet, initialized to :code:`None` and set to the right value if
        a facet is around.
    :ivar maxInclusive: Value for the :code:`xsd:maxInclusive` facet, initialized to :code:`None` and set to the right value if
        a facet is around.
    :ivar minLength: Value for the :code:`xsd:minLength` facet, initialized to :code:`None` and set to the right value if a facet
        is around.
    :ivar maxLength: Value for the :code:`xsd:maxLength` facet, initialized to :code:`None` and set to the right value if a facet
        is around.
    :ivar length: Value for the :code:`xsd:length` facet, initialized to :code:`None` and set to the right value if a facet is
        around.
    :ivar pattern: Array of patterns for the :code:`xsd:pattern` facet, initialized to :code:`[]` and set to the right value if a
        facet is around.
    :ivar langRange: Array of language ranges for the :code:`rdf:langRange` facet, initialized to :code:`[]` and set to the right
        value if a facet is around.
    :ivar check_methods: List of class methods that are relevant for the given :code:`base_type`.

    :ivar toPython: Function to convert a Literal of the specified type to a Python value. Is defined by :code:`lambda v:
        _lit_to_value(self, v)`, see :py:func:`._lit_to_value`.
    """

    def __init__(self, type_uri, base_type, facets):
        """
        @param type_uri: URI of the datatype being defined
        @param base_type: URI of the base datatype, ie, the one being restricted
        @param facets: array of C{(facetURI, value)} pairs
        """
        RestrictedDatatypeCore.__init__(self, type_uri, base_type)
        if self.base_type not in AltXSDToPYTHON:
            raise Exception("No facet is implemented for datatype %s" % self.base_type)
        self.converter = AltXSDToPYTHON[self.base_type]

        self.minExclusive = None
        self.maxExclusive = None
        self.minInclusive = None
        self.maxInclusive = None
        self.length = None
        self.maxLength = None
        self.minLength = None
        self.pattern = []
        self.langRange = []
        for (facet, value) in facets:
            if facet == XSD.minInclusive and (
                self.minInclusive is None or self.minInclusive < value
            ):
                self.minInclusive = value
            elif facet == XSD.minExclusive and (
                self.minExclusive is None or self.minExclusive < value
            ):
                self.minExclusive = value
            elif facet == XSD.maxInclusive and (
                self.maxInclusive is None or value < self.maxInclusive
            ):
                self.maxInclusive = value
            elif facet == XSD.maxExclusive and (
                self.maxExclusive is None or value < self.maxExclusive
            ):
                self.maxExclusive = value
            elif facet == RDF.langRange:
                self.langRange.append(value)
            elif facet == XSD.length:
                self.length = value
            elif facet == XSD.maxLength and (
                self.maxLength is None or value < self.maxLength
            ):
                self.maxLength = value
            elif facet == XSD.minLength and (
                self.minLength is None or value > self.minLength
            ):
                self.minLength = value
            elif facet == XSD.pattern:
                self.pattern.append(re.compile(value))

        # Choose the methods that are relevant for this datatype, based on the base type
        facet_to_method = {
            MIN_MAX: [
                RestrictedDatatype._check_max_exclusive,
                RestrictedDatatype._check_min_exclusive,
                RestrictedDatatype._check_max_inclusive,
                RestrictedDatatype._check_min_inclusive,
            ],
            LENGTH: [
                RestrictedDatatype._check_min_length,
                RestrictedDatatype._check_max_length,
                RestrictedDatatype._check_length,
            ],
            LENGTH_AND_PATTERN: [
                RestrictedDatatype._check_min_length,
                RestrictedDatatype._check_max_length,
                RestrictedDatatype._check_length,
                RestrictedDatatype._check_pattern,
            ],
            LENGTH_PATTERN_LRANGE: [
                RestrictedDatatype._check_min_length,
                RestrictedDatatype._check_max_length,
                RestrictedDatatype._check_length,
                RestrictedDatatype._check_lang_range,
            ],
        }
        self.check_methods = []
        for cat in Datatypes_per_facets:
            if self.base_type in Datatypes_per_facets[cat]:
                self.check_methods = facet_to_method[cat]
                break
        self.toPython = lambda v: _lit_to_value(self, v)

    def checkValue(self, value):
        """
        Check whether the (Python) value abides to the constraints defined by the current facets.

        :param value: The value to be checked.
        :rtype: bool
        """
        for method in self.check_methods:
            if not method(self, value):
                return False
        return True

    def _check_min_exclusive(self, value):
        """
        Check  the (python) value against min exclusive facet.
        @param value: the value to be checked
        @rtype: boolean
        """
        if self.minExclusive is not None:
            return self.minExclusive < value
        else:
            return True

    def _check_min_inclusive(self, value):
        """
        Check  the (python) value against min inclusive facet.
        @param value: the value to be checked
        @rtype: boolean
        """
        if self.minInclusive is not None:
            return self.minInclusive <= value
        else:
            return True

    def _check_max_exclusive(self, value):
        """
        Check  the (python) value against max exclusive facet.
        @param value: the value to be checked
        @rtype: boolean
        """
        if self.maxExclusive is not None:
            return value < self.maxExclusive
        else:
            return True

    def _check_max_inclusive(self, value):
        """
        Check  the (python) value against max inclusive facet.
        @param value: the value to be checked
        @rtype: boolean
        """
        if self.maxInclusive is not None:
            return value <= self.maxInclusive
        else:
            return True

    def _check_min_length(self, value):
        """
        Check  the (python) value against minimum length facet.
        @param value: the value to be checked
        @rtype: boolean
        """
        if isinstance(value, rdflibLiteral):
            val = str(value)
        else:
            val = value
        if self.minLength is not None:
            return self.minLength <= len(val)
        else:
            return True

    def _check_max_length(self, value):
        """
        Check  the (python) value against maximum length facet.
        @param value: the value to be checked
        @rtype: boolean
        """
        if isinstance(value, rdflibLiteral):
            val = str(value)
        else:
            val = value
        if self.maxLength is not None:
            return self.maxLength >= len(val)
        else:
            return True

    def _check_length(self, value):
        """
        Check  the (python) value against exact length facet.
        @param value: the value to be checked
        @rtype: boolean
        """
        if isinstance(value, rdflibLiteral):
            val = str(value)
        else:
            val = value
        if self.length is not None:
            return self.length == len(val)
        else:
            return True

    def _check_pattern(self, value):
        """
        Check  the (python) value against array of regular expressions.
        @param value: the value to be checked
        @rtype: boolean
        """
        if isinstance(value, rdflibLiteral):
            val = str(value)
        else:
            val = value
        for p in self.pattern:
            if p.match(val) is None:
                return False
        return True

    def _check_lang_range(self, value):
        """
        Check  the (python) value against array of language ranges.
        @param value: the value to be checked
        @rtype: boolean
        """
        if isinstance(value, rdflibLiteral):
            lang = value.language
        else:
            return False
        for r in self.langRange:
            if not _lang_range_check(r, lang):
                return False
        return True



# -*- coding: utf-8 -*-
#
"""
The generic superclasses for various rule based semantics and the possible extensions.

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""

__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from typing import Union

import rdflib
from rdflib.namespace import RDF
from rdflib import BNode, Literal, Graph, Dataset
from .Namespaces import ERRNS

try:
    from rdflib.graph import ConjunctiveGraph
except ImportError:
    ConjunctiveGraph = Dataset

debugGlobal = False
offlineGeneration = False


######################################################################################################
# noinspection PyMethodMayBeStatic,PyPep8Naming,PyPep8Naming
class Core:
    """
    Core of the semantics management, dealing with the RDFS and other Semantic triples. The only
    reason to have it in a separate class is for an easier maintainability.

    This is a common superclass only. In the present module, it is subclassed by
    a :class:`.RDFSClosure.RDFS_Semantics` class and a :class:`.OWLRL.OWLRL_Semantics` classes.
    There are some methods that are implemented in the subclasses only, ie, this class cannot be used by itself!

    :param graph: The RDF graph to be extended.
    :type graph: :class:`rdflib.graph.Graph`

    :param axioms: Whether axioms should be added or not.
    :type axioms: bool

    :param daxioms: Whether datatype axioms should be added or not.
    :type daxioms: bool

    :param rdfs: Whether RDFS inference is also done (used in subclassed only).
    :type rdfs: bool

    :param destination: The destination graph to which the results are written. If None, use the source graph.
    :type destination: :class:`rdflib.graph.Graph`

    :var IMaxNum: Maximal index of :code:`rdf:_i` occurrence in the graph.
    :type IMaxNum: int

    :var graph: The real graph.
    :type graph: :class:`rdflib.graph.Graph`

    :var axioms: Whether axioms should be added or not.
    :type axioms: bool

    :var daxioms: Whether datatype axioms should be added or not.
    :type daxioms: bool

    :var added_triples: Triples added to the graph, conceptually, during one processing cycle.
    :type added_triples: set of triples

    :var error_messages: Error messages (typically inconsistency messages in OWL RL) found during processing. These
        are added to the final graph at the very end as separate BNodes with error messages.
    :type error_messages: list of str

    :var rdfs: Whether RDFS inference is also done (used in subclassed only).
    :type rdfs: bool
    """

    # noinspection PyUnusedLocal
    def __init__(self, graph: Graph, axioms, daxioms, rdfs: bool = False, destination: Union[None, Graph] = None):
        """
        The parameter descriptions here are from the old documentation.

        @param graph: the RDF graph to be extended
        @type graph: Graph
        @param axioms: whether axioms should be added or not
        @type axioms: boolean
        @param daxioms: whether datatype axioms should be added or not
        @type daxioms: boolean
        @param rdfs: whether RDFS inference is also done (used in subclassed only)
        @type rdfs: boolean
        @param destination: the destination graph to which the results are written. If None, use the source graph.
        @type destination: Graph
        """
        self._debug = debugGlobal

        # Calculate the maximum 'n' value for the '_i' type predicates (see Horst's paper)
        n = 1
        maxnum = 0
        cont = True
        while cont:
            cont = False
            predicate = RDF[("_%d" % n)]
            for (s, p, o) in graph.triples((None, predicate, None)):
                # there is at least one if we got here
                maxnum = n
                n += 1
                cont = True
        self.IMaxNum = maxnum

        self.graph = graph
        if isinstance(self.graph, (Dataset, ConjunctiveGraph)):
            self.graph.default_union = True

        if destination is None:
            if isinstance(graph, (Dataset, ConjunctiveGraph)):
                self.destination = graph.default_context
            else:
                self.destination = graph
        else:
            if isinstance(destination, (str, rdflib.URIRef)):
                if isinstance(graph, (Dataset, ConjunctiveGraph)):
                    self.destination = graph.get_context(destination)
                else:
                    raise ValueError("URIRef destinations are only supported for Datasets and ConjunctiveGraphs")
            else:
                source_store = self.graph.store
                dest_store = destination.store
                if source_store is not dest_store:
                    raise ValueError("The source and destination graphs share the same backing store")
                self.destination = destination

        self.axioms = axioms
        self.daxioms = daxioms

        self.rdfs = rdfs

        self.error_messages = []
        self.empty_stored_triples()

    def add_error(self, message):
        """
        Add an error message

        :param message: Error message.
        :type message: str
        """
        if message not in self.error_messages:
            self.error_messages.append(message)

    def pre_process(self):
        """
        Do some pre-processing step. This method before anything else in the closure. By default, this method is empty,
        subclasses can add content to it by overriding it.
        """
        pass

    def post_process(self):
        """
        Do some post-processing step. This method when all processing is done, but before handling possible
        errors (ie, the method can add its own error messages). By default, this method is empty, subclasses
        can add content to it by overriding it.
        """
        pass

    def rules(self, t, cycle_num):
        """
        The core processing cycles through every tuple in the graph and dispatches it to the various methods
        implementing a specific group of rules. By default, this method raises an exception; indeed, subclasses
        *must* add content to by overriding it.

        :param t: One triple on which to apply the rules.
        :type t: tuple

        :param cycle_num: Which cycle are we in, starting with 1. This value is forwarded to all local rules; it is
            also used locally to collect the bnodes in the graph.
        :type cycle_num: int
        """
        raise Exception(
            "This method should not be called directly; subclasses should override it"
        )

    def add_axioms(self):
        """
        Add axioms.

        This is only a placeholder and raises an exception by default; subclasses *must* fill this with real content
        """
        raise Exception(
            "This method should not be called directly; subclasses should override it"
        )

    def add_d_axioms(self):
        """
        Add d axioms.

        This is only a placeholder and raises an exception by default; subclasses I{must} fill this with real content
        """
        raise Exception(
            "This method should not be called directly; subclasses should override it"
        )

    def one_time_rules(self):
        """
        This is only a placeholder; subclasses should fill this with real content. By default, it is just an empty call.
        This set of rules is invoked only once and not in a cycle.
        """
        pass

    # noinspection PyAttributeOutsideInit
    def empty_stored_triples(self):
        """
        Empty the internal store for triples.
        """
        self.added_triples = set()

    def flush_stored_triples(self):
        """
        Send the stored triples to the graph, and empty the container.
        """
        for t in self.added_triples:
            self.destination.add(t)
        self.empty_stored_triples()

    def store_triple(self, t):
        """
        In contrast to its name, this does not yet add anything to the graph itself, it just stores the tuple in an
        internal set (:code:`Core.added_triples`). (It is important for this to be a set: some of the rules in the
        various closures may generate the same tuples several times.) Before adding the tuple to the set, the method
        checks whether the tuple is in the final graph already (if yes, it is not added to the set).

        The set itself is emptied at the start of every processing cycle; the triples are then effectively added to the
        graph at the end of such a cycle. If the set is actually empty at that point, this means that the cycle has not
        added any new triple, and the full processing can stop.

        :param t: The triple to be added to the graph, unless it is already there
        :type t: tuple
        """
        (s, p, o) = t
        if not isinstance(p, Literal) and not (t in self.destination or t in self.graph):
            if self._debug or offlineGeneration:
                print(t)
            self.added_triples.add(t)

    # noinspection PyAttributeOutsideInit
    def closure(self):
        """
        Generate the closure the graph. This is the real 'core'.

        The processing rules store new triples via the separate method :func:`owlrl.Closure.Core.store_triple` which stores
        them in the :code:`added_triples` array. If that array is empty at the end of a cycle,
        it means that the whole process can be stopped.

        If required, the relevant axiomatic triples are added to the graph before processing in cycles. Similarly
        the exchange of literals against bnodes is also done in this step (and restored after all cycles are over).
        """
        self.pre_process()

        # Handling the axiomatic triples. In general, this means adding all tuples in the list that
        # forwarded, and those include RDF or RDFS. In both cases the relevant parts of the container axioms should also
        # be added.
        if self.axioms:
            self.add_axioms()

        # Add the datatype axioms, if needed (note that this makes use of the literal proxies, the order of the call
        # is important!
        if self.daxioms:
            self.add_d_axioms()

        self.flush_stored_triples()

        # Get first the 'one-time rules', ie, those that do not need an extra round in cycles down the line
        self.one_time_rules()
        self.flush_stored_triples()

        # Go cyclically through all rules until no change happens
        new_cycle = True
        cycle_num = 0
        while new_cycle:
            # yes, there was a change, let us go again
            cycle_num += 1

            # DEBUG: print the cycle number out
            if self._debug:
                print("----- Cycle #%d" % cycle_num)

            # go through all rules, and collect the replies (to see whether any change has been done)
            # the new triples to be added are collected separately not to interfere with
            # the current graph yet
            self.empty_stored_triples()

            # Execute all the rules; these might fill up the added triples array
            for t in self.graph.triples((None, None, None)):
                self.rules(t, cycle_num)

            # Add the tuples to the graph (if necessary, that is). If any new triple has been generated, a new cycle
            # will be necessary...
            new_cycle = len(self.added_triples) > 0

            for t in self.added_triples:
                self.destination.add(t)

        self.post_process()
        self.flush_stored_triples()

        # Add possible error messages
        if self.error_messages:
            # I am not sure this is the right vocabulary to use for this purpose, but I haven't found anything!
            # I could, of course, come up with my own, but I am not sure that would be kosher...
            self.destination.bind("err", "http://www.daml.org/2002/03/agents/agent-ont#")
            for m in self.error_messages:
                message = BNode()
                self.destination.add((message, RDF.type, ERRNS.ErrorMessage))
                self.destination.add((message, ERRNS.error, Literal(m)))



# -*- coding: utf-8 -*-
#
"""

Extension to OWL 2 RL, ie, some additional rules added to the system from OWL 2 Full. It is implemented through
the :class:`.OWLRL_Extension` class, whose reference has to be passed to the relevant semantic class (i.e., either the OWL 2 RL
or the combined closure class) as an 'extension'.

The added rules and features are:

    - self restriction
    - owl:rational datatype
    - datatype restrictions via facets
 
In more details, the rules that are added:

    1. self restriction 1: :code:`?z owl:hasSelf ?x. ?x owl:onProperty ?p. ?y rdf:type ?z. => ?y ?p ?y.`
    2. self restriction 2: :code:`?z owl:hasSelf ?x. ?x owl:onProperty ?p. ?y ?p ?y. => ?y rdf:type ?z.`

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""

__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from typing import Union

import rdflib
from rdflib import Graph
from rdflib.namespace import RDF, RDFS, OWL, XSD

from fractions import Fraction as Rational

from .DatatypeHandling import AltXSDToPYTHON


# noinspection PyPep8Naming
from .CombinedClosure import RDFS_OWLRL_Semantics
from .OWLRL import OWLRL_Annotation_properties

from .XsdDatatypes import OWL_RL_Datatypes, OWL_Datatype_Subsumptions

from .RestrictedDatatype import extract_faceted_datatypes

#######################################################################################################################
# Rational datatype


# noinspection PyPep8Naming
def _strToRational(v):
    """Converting a string to a rational.

    According to the OWL spec: numerator must be an integer, denominator a positive integer (ie, xsd['integer'] type),
    and the denominator should not have a '+' sign.

    @param v: the literal string defined as boolean
    @return corresponding Rational value
    @rtype: Rational
    @raise ValueError: invalid rational string literal
    """
    try:
        r = v.split("/")
        if len(r) == 2:
            n_str = r[0]
            d_str = r[1]
        else:
            n_str = r[0]
            d_str = "1"
        if d_str.strip()[0] == "+":
            raise ValueError("Invalid Rational literal value %s" % v)
        else:
            return Rational(
                AltXSDToPYTHON[XSD.integer](n_str),
                AltXSDToPYTHON[XSD.positiveInteger](d_str),
            )
    except:
        raise ValueError("Invalid Rational literal value %s" % v)


#######################################################################################################################


# noinspection PyPep8Naming,PyBroadException
class OWLRL_Extension(RDFS_OWLRL_Semantics):
    """
    Additional rules to OWL 2 RL. The initialization method also adds the :code:`owl:rational` datatype to the set of
    allowed datatypes with the :func:`owlrl.OWLRLExtras._strToRational` function as a conversion between the literal form and a Rational. The
    :code:`xsd:decimal` datatype is also set to be a subclass of :code:`owl:rational`. Furthermore, the restricted datatypes are
    extracted from the graph using a separate method in a different module
    (:py:func:`.RestrictedDatatype.extract_faceted_datatypes`), and all those datatypes are also added to the set of allowed
    datatypes. In the case of the restricted datatypes and extra subsumption relationship is set up between the
    restricted and the base datatypes.

    :cvar extra_axioms: Additional axioms that have to be added to the deductive closure (in case the axiomatic triples
        are required).

    :var restricted_datatypes: list of the datatype restriction from :class:`.RestrictedDatatype`.
    :type restricted_datatypes: list of L{restricted datatype<RestrictedDatatype.RestrictedDatatype>} instances
    """

    extra_axioms = [
        (OWL.hasSelf, RDF.type, RDF.Property),
        (OWL.hasSelf, RDFS.domain, RDF.Property),
    ]

    def __init__(self, graph: Graph, axioms, daxioms, rdfs: bool = False, destination: Union[None, Graph] = None):
        """
        @param graph: the RDF graph to be extended
        @type graph: rdflib.Graph
        @param axioms: whether (non-datatype) axiomatic triples should be added or not
        @type axioms: Boolean
        @param daxioms: whether datatype axiomatic triples should be added or not
        @type daxioms: Boolean
        @param rdfs: whether RDFS extension is done
        @type rdfs: boolean
        @param destination: the destination graph to which the results are written. If None, use the source graph.
        @type destination: rdflib.Graph
        """
        RDFS_OWLRL_Semantics.__init__(self, graph, axioms, daxioms, rdfs=rdfs, destination=destination)
        self.rdfs = rdfs
        self.add_new_datatype(
            OWL.rational,
            _strToRational,
            OWL_RL_Datatypes,
            subsumption_dict=OWL_Datatype_Subsumptions,
            subsumption_key=XSD.decimal,
            subsumption_list=[OWL.rational],
        )

        self.restricted_datatypes = extract_faceted_datatypes(self, graph)
        for dt in self.restricted_datatypes:
            self.add_new_datatype(
                dt.datatype,
                dt.toPython,
                OWL_RL_Datatypes,
                subsumption_dict=OWL_Datatype_Subsumptions,
                subsumption_key=dt.datatype,
                subsumption_list=[dt.base_type],
            )

    # noinspection PyShadowingNames
    def _subsume_restricted_datatypes(self):
        """
        A one-time-rule: all the literals are checked whether they are (a) of type restricted by a
        faceted (restricted) datatype and (b) whether
        the corresponding value abides to the restrictions. If true, then the literal gets an extra
        tag as being of type of the restricted datatype, too.
        """
        literals = self._literals()
        for rt in self.restricted_datatypes:
            # This is a recorded restriction. The base type is:
            base_type = rt.base_type
            # Look through all the literals
            for lt in literals:
                # check if the type of that literal matches. Note that this also takes
                # into account the subsumption datatypes, that have been taken
                # care of by the 'regular' OWL RL process
                if (lt, RDF.type, base_type) in self.graph:
                    try:
                        # the conversion or the check may go wrong and raise an exception; then simply move on
                        if rt.checkValue(lt.toPython()):
                            # yep, this is also of type 'rt'
                            self.store_triple((lt, RDF.type, rt.datatype))
                    except:
                        continue

    def restriction_typing_check(self, v, t):
        """
        Helping method to check whether a type statement is in line with a possible
        restriction. This method is invoked by rule "cls-avf" before setting a type
        on an allValuesFrom restriction.

        The method is a placeholder at this level. It is typically implemented by subclasses for
        extra checks, e.g., for datatype facet checks.

        :param v: the resource that is to be 'typed'.
        :param t: the targeted type (i.e., Class).
        :return: Boolean.
        :rtype: bool
        """
        # Look through the restricted datatypes to see if 't' corresponds to one of those...
        # There are a bunch of possible exceptions here with datatypes, but they can all
        # be ignored...
        try:
            for rt in self.restricted_datatypes:
                if rt.datatype == t:
                    # bingo
                    if v in self.literal_proxies.bnode_to_lit:
                        return rt.checkValue(
                            self.literal_proxies.bnode_to_lit[v].lit.toPython()
                        )
                    else:
                        return True
            # if we got here, no restriction applies
            return True
        except:
            return True

    def one_time_rules(self):
        """
        This method is invoked only once at the beginning, and prior of, the forward chaining process.

        At present, only the L{subsumption} of restricted datatypes<_subsume_restricted_datatypes>} is performed.
        """
        RDFS_OWLRL_Semantics.one_time_rules(self)
        # it is important to flush the triples at this point, because
        # the handling of the restriction datatypes rely on the datatype
        # subsumption triples added by the superclass
        self.flush_stored_triples()
        self._subsume_restricted_datatypes()

    def add_axioms(self):
        """
        Add the :class:`owlrl.OWLRLExtras.OWLRL_Extension.extra_axioms`, related to the self restrictions. This method is invoked only once at the beginning, and prior of, the forward chaining process.
        """
        RDFS_OWLRL_Semantics.add_axioms(self)
        for t in self.extra_axioms:
            self.destination.add(t)

    def rules(self, t, cycle_num):
        """
        Go through the additional rules implemented by this module.

        :param t: A triple (in the form of a tuple).
        :type t: tuple

        :param cycle_num: Which cycle are we in, starting with 1. This value is forwarded to all local rules; it is
            also used locally to collect the bnodes in the graph.
        :type cycle_num: int
        """
        RDFS_OWLRL_Semantics.rules(self, t, cycle_num)
        z, q, x = t
        if q == OWL.hasSelf:
            for p in self.graph.objects(z, OWL.onProperty):
                for y in self.graph.subjects(RDF.type, z):
                    self.store_triple((y, p, y))
                for y1, y2 in self.graph.subject_objects(p):
                    if y1 == y2:
                        self.store_triple((y1, RDF.type, z))


# noinspection PyPep8Naming
class OWLRL_Extension_Trimming(OWLRL_Extension):
    """
    This Class adds only one feature to :class:`.OWLRL_Extension`: to initialize with a trimming flag set to :code:`True` by
    default.

    This is pretty experimental and probably contentious: this class *removes* a number of triples from the Graph at
    the very end of the processing steps. These triples are either the by-products of the deductive closure calculation
    or are axiom like triples that are added following the rules of OWL 2 RL. While these triples *are necessary* for
    the correct inference of really 'useful' triples, they may not be of interest for the application for the end
    result. The triples that are removed are of the form (following a SPARQL-like notation):

    - :code:`?x owl:sameAs ?x`, :code:`?x rdfs:subClassOf ?x`, :code:`?x rdfs:subPropertyOf ?x`, :code:`?x owl:equivalentClass ?x` type triples.

    - :code:`?x rdfs:subClassOf rdf:Resource`, :code:`?x rdfs:subClassOf owl:Thing`, :code:`?x rdf:type rdf:Resource`, :code:`owl:Nothing rdfs:subClassOf ?x` type triples.

    - For a datatype that does *not* appear explicitly in a type assignments (ie, in a :code:`?x rdf:type dt`) the corresponding :code:`dt rdf:type owl:Datatype` and :code:`dt rdf:type owl:DataRange` triples, as well as the disjointness statements with other datatypes.
    - annotation property axioms.
    - a number of axiomatic triples on :code:`owl:Thing`, :code:`owl:Nothing` and :code:`rdf:Resource` (eg, :code:`owl:Nothing rdf:type owl:Class`, :code:`owl:Thing owl:equivalentClass rdf:Resource`, etc).

    Trimming is the only feature of this class, done in the :py:meth:`.post_process` step. If users extend :class:`OWLRL_Extension`,
    this class can be safely mixed in via multiple inheritance.

    :param graph: The RDF graph to be extended.
    :type graph: :class:`rdflib.Graph`

    :param axioms: Whether (non-datatype) axiomatic triples should be added or not.
    :type axioms: bool

    :param daxioms: Whether datatype axiomatic triples should be added or not.
    :type daxioms: bool

    :param rdfs: Whether RDFS extension is done.
    :type rdfs: bool
    """

    def __init__(self, graph: Graph, axioms, daxioms, rdfs: bool = False, destination: Union[None, Graph] = None):
        """
        @param graph: the RDF graph to be extended
        @type graph: rdflib.Graph
        @param axioms: whether (non-datatype) axiomatic triples should be added or not
        @type axioms: Boolean
        @param daxioms: whether datatype axiomatic triples should be added or not
        @type daxioms: Boolean
        @param rdfs: whether RDFS extension is done
        @type rdfs: boolean
        @param destination: the destination graph to which the results are written. If None, use the source graph.
        @type destination: rdflib.Graph
        """
        OWLRL_Extension.__init__(self, graph, axioms, daxioms, rdfs=rdfs, destination=destination)

    def post_process(self):
        """
        Do some post-processing step performing the trimming of the graph. See the :class:`.OWLRL_Extension_Trimming`
        class for further details.
        """
        OWLRL_Extension.post_process(self)
        self.flush_stored_triples()

        to_be_removed = set()
        for t in self.graph.triples((None, None, None)):
            s, p, o = t
            if s == o:
                if (
                    p == OWL.sameAs
                    or p == OWL.equivalentClass
                    or p == RDFS.subClassOf
                    or p == RDFS.subPropertyOf
                ):
                    to_be_removed.add(t)
            if (
                (p == RDFS.subClassOf and (o == OWL.Thing or o == RDFS.Resource))
                or (p == RDF.type and o == RDFS.Resource)
                or (s == OWL.Nothing and p == RDFS.subClassOf)
            ):
                to_be_removed.add(t)

        for dt in OWL_RL_Datatypes:
            # see if this datatype appears explicitly in the graph as the type of a symbol
            if len([s for s in self.graph.subjects(RDF.type, dt)]) == 0:
                to_be_removed.add((dt, RDF.type, RDFS.Datatype))
                to_be_removed.add((dt, RDF.type, OWL.DataRange))

                for t in self.graph.triples((dt, OWL.disjointWith, None)):
                    to_be_removed.add(t)
                for t in self.graph.triples((None, OWL.disjointWith, dt)):
                    to_be_removed.add(t)

        for an in OWLRL_Annotation_properties:
            self.destination.remove((an, RDF.type, OWL.AnnotationProperty))

        to_be_removed.add((OWL.Nothing, RDF.type, OWL.Class))
        to_be_removed.add((OWL.Nothing, RDF.type, RDFS.Class))
        to_be_removed.add((OWL.Thing, RDF.type, OWL.Class))
        to_be_removed.add((OWL.Thing, RDF.type, RDFS.Class))
        to_be_removed.add((OWL.Thing, OWL.equivalentClass, RDFS.Resource))
        to_be_removed.add((RDFS.Resource, OWL.equivalentClass, OWL.Thing))
        to_be_removed.add((OWL.Class, OWL.equivalentClass, RDFS.Class))
        to_be_removed.add((OWL.Class, RDFS.subClassOf, RDFS.Class))
        to_be_removed.add((RDFS.Class, OWL.equivalentClass, OWL.Class))
        to_be_removed.add((RDFS.Class, RDFS.subClassOf, OWL.Class))
        to_be_removed.add((RDFS.Datatype, RDFS.subClassOf, OWL.DataRange))
        to_be_removed.add((RDFS.Datatype, OWL.equivalentClass, OWL.DataRange))
        to_be_removed.add((OWL.DataRange, RDFS.subClassOf, OWL.DatatypeProperty))
        to_be_removed.add((OWL.DataRange, OWL.equivalentClass, OWL.DatatypeProperty))

        for t in to_be_removed:
            self.destination.remove(t)



# -*- coding: utf-8 -*-
#
"""
This module is a **brute force** implementation of the OWL 2 RL profile.

RDFLib works with 'generalized' RDF, meaning that triples with a BNode predicate are *allowed*. This is good because,
e.g., some of the triples generated for RDF from an OWL 2 functional syntax might look like :code:`[ owl:inverseOf p ]`, and the
RL rules would then operate on such generalized triple. However, as a last, post processing steps, these triples are 
removed from the graph before serialization to produce 'standard' RDF (which is o.k. for RL, too, because the 
consequent triples are all right, generalized triples might have had a role in the deduction steps only).

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""

__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from collections import defaultdict
from typing import Union

import rdflib
from rdflib import BNode, Graph
from rdflib.namespace import OWL, RDF, RDFS

from owlrl.Closure import Core
from owlrl.AxiomaticTriples import OWLRL_Axiomatic_Triples, OWLRL_D_Axiomatic_Triples
from owlrl.AxiomaticTriples import OWLRL_Datatypes_Disjointness

OWLRL_Annotation_properties = [
    RDFS.label,
    RDFS.comment,
    RDFS.seeAlso,
    RDFS.isDefinedBy,
    OWL.deprecated,
    OWL.versionInfo,
    OWL.priorVersion,
    OWL.backwardCompatibleWith,
    OWL.incompatibleWith,
]

from .XsdDatatypes import OWL_RL_Datatypes, OWL_Datatype_Subsumptions
from .DatatypeHandling import AltXSDToPYTHON


identity = lambda v: v


#######################################################################################################################


# OWL-R Semantics class
#
#
# As an editing help: each rule is prefixed by RULE XXXX where XXXX is the acronym given in the profile document.
# This helps in referring back to the spec...
# noinspection PyPep8Naming, PyPep8Naming, PyBroadException
class OWLRL_Semantics(Core):
    """
    OWL 2 RL Semantics class, i.e., implementation of the OWL 2 RL closure graph.

    .. note:: Note that the module does *not* implement the so called Datatype entailment rules, simply because the underlying
        RDFLib does not implement the datatypes (i.e., RDFLib will not make the literal "1.00" and "1.00000" identical,
        although even with all the ambiguities on datatypes, this *should* be made equal...).

        Also, the so-called extensional entailment rules (Section 7.3.1 in the RDF Semantics document) have not been
        implemented either.

    The comments and references to the various rule follow the names as used in the `OWL 2 RL
    document`_.

    .. _OWL 2 RL document: http://www.w3.org/TR/owl2-profiles/#Reasoning_in_OWL_2_RL_and_RDF_Graphs_using_Rules

    :param graph: The RDF graph to be extended.
    :type graph: :class:`rdflib.Graph`

    :param axioms: Whether (non-datatype) axiomatic triples should be added or not.
    :type axioms: bool

    :param daxioms: Whether datatype axiomatic triples should be added or not.
    :type daxioms: bool

    :param rdfs: Whether RDFS inference is also done (used in subclassed only).
    :type rdfs: bool
    """

    def __init__(self, graph: Graph, axioms, daxioms, rdfs: bool = False, destination: Union[None, Graph] = None):
        """
        @param graph: the RDF graph to be extended
        @type graph: rdflib.Graph
        @param axioms: whether (non-datatype) axiomatic triples should be added or not
        @type axioms: bool
        @param daxioms: whether datatype axiomatic triples should be added or not
        @type daxioms: bool
        @param rdfs: whether RDFS inference is also done (used in subclassed only)
        @type rdfs: boolean
        @param destination: the destination graph to which the results are written. If None, use the source graph.
        @type destination: rdflib.Graph
        """
        Core.__init__(self, graph, axioms, daxioms, rdfs=rdfs, destination=destination)
        self.bnodes = []

    def _list(self, l):
        """
        Shorthand to get a list of values (ie, from an rdf:List structure) starting at a head

        @param l: RDFLib resource, should be the head of an rdf:List
        @return: array of resources
        """
        return [ch for ch in self.graph.items(l)]

    def post_process(self):
        """
        Remove triples with Bnode predicates. The Bnodes in the graph are collected in the first cycle run.
        """
        to_be_removed = []
        for b in self.bnodes:
            for t in self.graph.triples((None, b, None)):
                if t not in to_be_removed:
                    to_be_removed.append(t)
        for t in to_be_removed:
            self.destination.remove(t)
            self.graph.remove(t)

    def add_axioms(self):
        """
        Add axioms
        """
        for t in OWLRL_Axiomatic_Triples:
            self.destination.add(t)

    def add_d_axioms(self):
        """
        Add the datatype axioms
        """
        for t in OWLRL_D_Axiomatic_Triples:
            self.destination.add(t)

    def restriction_typing_check(self, v, t):
        """
        Helping method to check whether a type statement is in line with a possible
        restriction. This method is invoked by rule "cls-avf" before setting a type
        on an allValuesFrom restriction.

        The method is a placeholder at this level. It is typically implemented by subclasses for
        extra checks, e.g., for datatype facet checks.

        :param v: The resource that is to be 'typed'.
        :param t: The targeted type (ie, Class).
        :return: Boolean.
        :rtype: bool
        """
        return True

    def _one_time_rules_datatypes(self):
        """
        Some of the rules in the rule set are axiomatic in nature, meaning that they really have to be added only
        once, there is no reason to add these in a cycle. These are performed by this method that is invoked only once
        at the beginning of the process.

        These are: cls-thing, cls-nothing1, prp-ap, dt-types1, dt-types2, dt-eq, dt-diff.

        .. note:: Note, however, that the dt-* are executed only partially, limited by the possibilities offered by RDFLib. These
            may not cover all the edge cases of OWL RL. Especially, dt-not-type has not (yet?) been implemented (I wonder
            whether RDFLib should not raise exception for those anyway...
        """
        # noinspection PyShadowingNames
        def _add_to_explicit(s, o):
            explicit[s].add(o)

        # noinspection PyShadowingNames
        def _append_to_explicit(s, o):
            explicit[s].add(o)

        # noinspection PyShadowingNames
        def _handle_subsumptions(r, dt):
            if dt in OWL_Datatype_Subsumptions:
                for new_dt in OWL_Datatype_Subsumptions[dt]:
                    self.store_triple((r, RDF.type, new_dt))
                    self.store_triple((new_dt, RDF.type, RDFS.Datatype))
                    used_datatypes.add(new_dt)

        # explicit object->datatype relationships: those that came from an object being typed as a datatype
        # or a sameAs. The values are arrays of datatypes to which the resource belong
        explicit = defaultdict(set)

        # For processing later:
        # implicit object->datatype relationships: these come from real
        # literals which are present in the graph
        implicit = {
            o: o.datatype
            for s, p, o in self.graph.triples((None, None, None))
            if isinstance(o, rdflib.Literal) and o.datatype in OWL_RL_Datatypes
        }

        # datatypes in use by the graph (directly or indirectly). This will be used at the end to add the
        # necessary disjointness statements (but not more)
        used_datatypes = set(implicit.values())

        # RULE dt-type2: for all explicit literals the corresponding bnode should get the right type
        # definition. The 'implicit' dictionary is also filled on the fly
        # RULE dt-not-type: see whether an explicit literal is valid in terms of the defined datatype
        for lt in implicit:  # note that all non-RL datatypes are ignored
            # add the explicit typing triple
            self.store_triple((lt, RDF.type, lt.datatype))

            # for dt-not-type
            # This is a dirty trick: rdflib's Literal includes a method that raises an exception if the
            # lexical value cannot be mapped on the value space.
            converter = AltXSDToPYTHON.get(lt.datatype, identity)
            try:
                converter(str(lt))
            except ValueError:
                self.add_error(
                    "Lexical value of the literal '%s' does not match"
                    " its datatype (%s)" % (lt, lt.datatype)
                )

        # RULE dt-diff
        # RULE dt-eq
        # Compare literals whether they are different or not. This rules
        # are skipped on purpose at the moment.

        # Other datatype definitions can come from explicitly defining some nodes as datatypes (though rarely used,
        # it is perfectly possible...
        # there may be explicit relationships set in the graph, too!
        for (s, p, o) in self.graph.triples((None, RDF.type, None)):
            if o in OWL_RL_Datatypes:
                used_datatypes.add(o)
                if s not in implicit:
                    _add_to_explicit(s, o)

        # Finally, there may be sameAs statements that bind nodes to some of the existing ones. This does not introduce
        # new datatypes, so the used_datatypes array does not get extended
        for (s, p, o) in self.graph.triples((None, OWL.sameAs, None)):
            if o in implicit:
                _add_to_explicit(s, implicit[o])
            # note that s in implicit would mean that the original graph has
            # a literal in subject position which is not allowed at the moment, so I do not bother
            if o in explicit:
                _append_to_explicit(s, o)
            if s in explicit:
                _append_to_explicit(o, s)

        # what we have now:
        # explicit+implicit contains all the resources of type datatype;
        # implicit contains those that are given by an explicit literal
        # explicit contains those that are given by general resources, and there can be a whole array for each entry

        # RULE dt-type1: add a Datatype typing for all those
        # Note: the strict interpretation of OWL RL is to do that for all allowed datatypes, but this is
        # under discussion right now. The optimized version uses only what is really in use
        for dt in OWL_RL_Datatypes:
            self.store_triple((dt, RDF.type, RDFS.Datatype))
        for dts in explicit.values():
            for dt in dts:
                self.store_triple((dt, RDF.type, RDFS.Datatype))

        # Datatype reasoning means that certain datatypes are subtypes of one another.
        # I could simply generate the extra subclass relationships into the graph and let the generic
        # process work its way, but it seems to be an overkill. Instead, I prefer to add the explicit typing
        # into the graph 'manually'
        for r in explicit:
            # these are the datatypes that this resource has
            dtypes = explicit[r]
            for dt in dtypes:
                _handle_subsumptions(r, dt)

        for r, dt in implicit.items():
            _handle_subsumptions(r, dt)

        # Last step: add the datatype disjointness relationships. This is done only for
        #  - 'top' level datatypes
        #  - used in the graph
        for t in OWLRL_Datatypes_Disjointness:
            (l, pred, r) = t
            if l in used_datatypes and r in used_datatypes:
                self.store_triple(t)

    def _one_time_rules_misc(self):
        """
        Rules executed: cls-thing, cls-nothing, prp-ap.
        """
        # RULE cls-thing
        self.store_triple((OWL.Thing, RDF.type, OWL.Class))

        # RULE cls-nothing
        self.store_triple((OWL.Nothing, RDF.type, OWL.Class))

        # RULE prp-ap
        for an in OWLRL_Annotation_properties:
            self.store_triple((an, RDF.type, OWL.AnnotationProperty))

    def one_time_rules(self):
        """
        Some of the rules in the rule set are axiomatic in nature, meaning that they really have to be added only
        once, there is no reason to add these in a cycle. These are performed by this method that is invoked only once
        at the beginning of the process.

        These are: cls-thing, cls-nothing1, prp-ap, dt-types1, dt-types2, dt-eq, dt-diff.
        """
        self._one_time_rules_misc()
        self._one_time_rules_datatypes()

    def rules(self, t, cycle_num):
        """
        Go through the various rule groups, as defined in the OWL-RL profile text and implemented via
        local methods. (The calling cycle takes every tuple in the graph.)

        :param t: A triple (in the form of a tuple).
        :param cycle_num: Which cycle are we in, starting with 1. This value is forwarded to all local rules; it is
            also used locally to collect the bnodes in the graph.
        """
        if cycle_num == 1:
            for r in t:
                if isinstance(r, BNode) and r not in self.bnodes:
                    self.bnodes.append(r)

        self._properties(t, cycle_num)
        self._equality(t, cycle_num)
        self._classes(t, cycle_num)
        self._class_axioms(t, cycle_num)
        self._schema_vocabulary(t, cycle_num)

    def _property_chain(self, p, x):
        """
        Implementation of the property chain axiom, invoked from inside the property axiom handler. This is the
        implementation of rule prp-spo2, taken aside for an easier readability of the code."""
        chain = self._list(x)
        # The complication is that, at each step of the chain, there may be spawns, leading to a multitude
        # of 'sub' chains:-(
        if len(chain) > 0:
            for (u1, _y, _z) in self.graph.triples((None, chain[0], None)):
                # At least the chain can be started, because the leftmost property has at least
                # one element in its extension
                finalList = [(u1, _z)]
                chainExists = True
                for pi in chain[1:]:
                    newList = []
                    for (_u, ui) in finalList:
                        for u in self.graph.objects(ui, pi):
                            # what is stored is only last entry with u1, the intermediate results
                            # are not of interest
                            newList.append((u1, u))
                    # I have now, in new list, all the intermediate results
                    # until pi
                    # if new list is empty, that means that is a blind alley
                    if len(newList) == 0:
                        chainExists = False
                        break
                    else:
                        finalList = newList
                if chainExists:
                    for (_u, un) in finalList:
                        self.store_triple((u1, p, un))

    def _equality(self, triple, cycle_num):
        """
        Table 4: Semantics of equality. Essentially, the eq-* rules.
        @param triple: triple to work on
        @param cycle_num: which cycle are we in, starting with 1. Can be used for some optimization.
        """
        # In many of the 'if' branches, corresponding to rules in the document,
        # the branch begins by a renaming of variables (eg, pp, c = s, o).
        # There is no programming reasons for doing that, but by renaming the
        # variables it becomes easier to compare the declarative rules
        # in the document with the implementation
        s, p, o = triple
        # RULE eq-ref
        self.store_triple((s, OWL.sameAs, s))
        self.store_triple((o, OWL.sameAs, o))
        self.store_triple((p, OWL.sameAs, p))
        if p == OWL.sameAs:
            x, y = s, o
            # RULE eq-sym
            self.store_triple((y, OWL.sameAs, x))
            # RULE eq-trans
            for z in self.graph.objects(y, OWL.sameAs):
                self.store_triple((x, OWL.sameAs, z))
            # RULE eq-rep-s
            for pp, oo in self.graph.predicate_objects(s):
                self.store_triple((o, pp, oo))
            # RULE eq-rep-p
            for ss, oo in self.graph.subject_objects(s):
                self.store_triple((ss, o, oo))
            # RULE eq-rep-o
            for ss, pp in self.graph.subject_predicates(o):
                self.store_triple((ss, pp, s))
            # RULE eq-diff1
            if (s, OWL.differentFrom, o) in self.graph or (
                o,
                OWL.differentFrom,
                s,
            ) in self.graph:
                self.add_error(
                    "'sameAs' and 'differentFrom' cannot be used on the same subject-object pair: (%s, %s)"
                    % (s, o)
                )

        # RULES eq-diff2 and eq-diff3
        if p == RDF.type and o == OWL.AllDifferent:
            x = s
            # the objects method are generators, we cannot simply concatenate them. So we turn the results
            # into lists first. (Otherwise the body of the for loops should be repeated verbatim, which
            # is silly and error prone...
            m1 = [i for i in self.graph.objects(x, OWL.members)]
            m2 = [i for i in self.graph.objects(x, OWL.distinctMembers)]
            for y in m1 + m2:
                zis = self._list(y)
                for i in range(0, len(zis) - 1):
                    zi = zis[i]
                    for j in range(i + 1, len(zis) - 1):
                        zj = zis[j]
                        if (
                            (zi, OWL.sameAs, zj) in self.graph
                            or (zj, OWL.sameAs, zi) in self.graph
                        ) and zi != zj:
                            self.add_error(
                                "'sameAs' and 'AllDifferent' cannot be used on the same subject-object "
                                "pair: (%s, %s)" % (zi, zj)
                            )

    def _properties(self, triple, cycle_num):
        """
        Table 5: The Semantics of Axioms about Properties. Essentially, the prp-* rules.
        @param triple: triple to work on
        @param cycle_num: which cycle are we in, starting with 1. Can be used for some optimization.
        """
        # In many of the 'if' branches, corresponding to rules in the document,
        # the branch begins by a renaming of variables (eg, pp, c = s, o).
        # There is no programming reasons for doing that, but by renaming the
        # variables it becomes easier to compare the declarative rules
        # in the document with the implementation
        p, t, o = triple

        # RULE prp-ap
        if cycle_num == 1 and t in OWLRL_Annotation_properties:
            self.store_triple((t, RDF.type, OWL.AnnotationProperty))

        # RULE prp-dom
        if t == RDFS.domain:
            for x, y in self.graph.subject_objects(p):
                self.store_triple((x, RDF.type, o))

        # RULE prp-rng
        elif t == RDFS.range:
            for x, y in self.graph.subject_objects(p):
                self.store_triple((y, RDF.type, o))

        elif t == RDF.type:
            # RULE prp-fp
            if o == OWL.FunctionalProperty:
                # Property axiom #3
                for x, y1 in self.graph.subject_objects(p):
                    for y2 in self.graph.objects(x, p):
                        # Optimization: if the two resources are identical, the samAs is already
                        # taken place somewhere else, unnecessary to add it here
                        if y1 != y2:
                            self.store_triple((y1, OWL.sameAs, y2))

            # RULE prp-ifp
            elif o == OWL.InverseFunctionalProperty:
                for x1, y in self.graph.subject_objects(p):
                    for x2 in self.graph.subjects(p, y):
                        # Optimization: if the two resources are identical, the samAs is already
                        # taken place somewhere else, unnecessary to add it here
                        if x1 != x2:
                            self.store_triple((x1, OWL.sameAs, x2))

            # RULE prp-irp
            elif o == OWL.IrreflexiveProperty:
                for x, y in self.graph.subject_objects(p):
                    if x == y:
                        self.add_error(
                            "Irreflexive property used on %s with %s" % (x, p)
                        )

            # RULE prp-symp
            elif o == OWL.SymmetricProperty:
                for x, y in self.graph.subject_objects(p):
                    self.store_triple((y, p, x))

            # RULE prp-asyp
            elif o == OWL.AsymmetricProperty:
                for x, y in self.graph.subject_objects(p):
                    if (y, p, x) in self.graph:
                        self.add_error(
                            "Erroneous usage of asymmetric property %s on %s and %s"
                            % (p, x, y)
                        )

            # RULE prp-trp
            elif o == OWL.TransitiveProperty:
                for x, y in self.graph.subject_objects(p):
                    for z in self.graph.objects(y, p):
                        self.store_triple((x, p, z))

            #
            # Breaking the order here, I take some additional rules here to the branch checking the type...
            #
            # RULE prp-adp
            elif o == OWL.AllDisjointProperties:
                x = p
                for y in self.graph.objects(x, OWL.members):
                    pis = self._list(y)
                    for i in range(0, len(pis) - 1):
                        pi = pis[i]
                        for j in range(i + 1, len(pis) - 1):
                            pj = pis[j]
                            for x, y in self.graph.subject_objects(pi):
                                if (x, pj, y) in self.graph:
                                    self.add_error(
                                        "Disjoint properties in an 'AllDisjointProperties' are not really "
                                        "disjoint: (%s, %s,%s) and (%s,%s,%s)"
                                        % (x, pi, y, x, pj, y)
                                    )

        # RULE prp-spo1
        elif t == RDFS.subPropertyOf:
            p1, p2 = p, o
            for x, y in self.graph.subject_objects(p1):
                self.store_triple((x, p2, y))

        # RULE prp-spo2
        elif t == OWL.propertyChainAxiom:
            self._property_chain(p, o)

        # RULES prp-eqp1 and prp-eqp2
        elif t == OWL.equivalentProperty:
            p1, p2 = p, o
            # Optimization: it clearly does not make sense to run these
            # if the two properties are identical (a separate axiom
            # does create an equivalent property relations among identical
            # properties, too...)
            if p1 != p2:
                # RULE prp-eqp1
                for x, y in self.graph.subject_objects(p1):
                    self.store_triple((x, p2, y))
                # RULE prp-eqp2
                for x, y in self.graph.subject_objects(p2):
                    self.store_triple((x, p1, y))

        # RULE prp-pdw
        elif t == OWL.propertyDisjointWith:
            p1, p2 = p, o
            for x, y in self.graph.subject_objects(p1):
                if (x, p2, y) in self.graph:
                    self.add_error(
                        "Erroneous usage of disjoint properties %s and %s on %s and %s"
                        % (p1, p2, x, y)
                    )

        # RULES prp-inv1 and prp-inv2
        elif t == OWL.inverseOf:
            p1, p2 = p, o
            # RULE prp-inv1
            for x, y in self.graph.subject_objects(p1):
                self.store_triple((y, p2, x))
            # RULE prp-inv2
            for x, y in self.graph.subject_objects(p2):
                self.store_triple((y, p1, x))

        # RULE prp-key
        elif t == OWL.hasKey:
            c, u = p, o
            pis = self._list(u)
            if len(pis) > 0:
                for x in self.graph.subjects(RDF.type, c):
                    # "Calculate" the keys for 'x'. The complication is that there can be various combinations
                    # of the keys, and that is the structure one has to build up here...
                    #
                    # The final list will be a list of lists, with each constituents being the possible combinations
                    # of the key values.
                    # startup the list
                    finalList = [[zi] for zi in self.graph.objects(x, pis[0])]
                    for pi in pis[1:]:
                        newList = []
                        for zi in self.graph.objects(x, pi):
                            newList = newList + [l + [zi] for l in finalList]
                        finalList = newList

                    # I am not sure this can happen, but better safe then sorry... ruling out
                    # the value lists whose length are not kosher
                    # (To be checked whether this is necessary in the first place)
                    valueList = [l for l in finalList if len(l) == len(pis)]

                    # Now we can look for the y-s, to see if they have the same key values
                    for y in self.graph.subjects(RDF.type, c):
                        # rule out the existing equivalences
                        if not (
                            y == x
                            or (y, OWL.sameAs, x) in self.graph
                            or (x, OWL.sameAs, y) in self.graph
                        ):
                            # 'calculate' the keys for the y values and see if there is a match
                            for vals in valueList:
                                same = True
                                for i in range(0, len(pis) - 1):
                                    if (y, pis[i], vals[i]) not in self.graph:
                                        same = False
                                        # No use going with this property line
                                        break
                                if same:
                                    self.store_triple((x, OWL.sameAs, y))
                                    # Look for the next 'y', this branch is finished, no reason to continue
                                    break

        # RULES prp-npa1 and prp-npa2
        elif t == OWL.sourceIndividual:
            x, i1 = p, o
            for p1 in self.graph.objects(x, OWL.assertionProperty):
                for i2 in self.graph.objects(x, OWL.targetIndividual):
                    if (i1, p1, i2) in self.graph:
                        self.add_error(
                            "Negative (object) property assertion violated for: (%s, %s, %s)"
                            % (i1, p1, i2)
                        )
                for i2 in self.graph.objects(x, OWL.targetValue):
                    if (i1, p1, i2) in self.graph:
                        self.add_error(
                            "Negative (datatype) property assertion violated for: (%s, %s, %s)"
                            % (i1, p1, i2)
                        )

    def _classes(self, triple, cycle_num):
        """
        Table 6: The Semantics of Classes. Essentially, the cls-* rules
        @param triple: triple to work on
        @param cycle_num: which cycle are we in, starting with 1. Can be used for some optimization.
        """
        # In many of the 'if' branches, corresponding to rules in the document,
        # the branch begins by a renaming of variables (eg, pp, c = s, o).
        # There is no programming reasons for doing that, but by renaming the
        # variables it becomes easier to compare the declarative rules
        # in the document with the implementation
        c, p, x = triple

        # RULE cls-nothing2
        if p == RDF.type and x == OWL.Nothing:
            self.add_error("%s is defined of type 'Nothing'" % c)

        # RULES cls-int1 and cls-int2
        if p == OWL.intersectionOf:
            classes = self._list(x)
            # RULE cls-int1
            # Optimization: by looking at the members of class[0] right away one
            # reduces the search spaces a bit. Individuals not in that class
            # are without interest anyway
            # I am not sure how empty lists are sanctioned, so having an extra check
            # on that does not hurt..
            if len(classes) > 0:
                for y in self.graph.subjects(RDF.type, classes[0]):
                    if False not in [
                        (y, RDF.type, cl) in self.graph for cl in classes[1:]
                    ]:
                        self.store_triple((y, RDF.type, c))
            # RULE cls-int2
            for y in self.graph.subjects(RDF.type, c):
                for cl in classes:
                    self.store_triple((y, RDF.type, cl))

        # RULE cls-uni
        elif p == OWL.unionOf:
            for cl in self._list(x):
                for y in self.graph.subjects(RDF.type, cl):
                    self.store_triple((y, RDF.type, c))

        # RULE cls-comm
        elif p == OWL.complementOf:
            c1, c2 = c, x
            for x1 in self.graph.subjects(RDF.type, c1):
                if (x1, RDF.type, c2) in self.graph:
                    self.add_error(
                        "Violation of complementarity for classes %s and %s on element %s"
                        % (c1, c2, x)
                    )

        # RULES cls-svf1 and cls=svf2
        elif p == OWL.someValuesFrom:
            xx, y = c, x
            # RULE cls-svf1
            # RULE cls-svf2
            for pp in self.graph.objects(xx, OWL.onProperty):
                for u, v in self.graph.subject_objects(pp):
                    if y == OWL.Thing or (v, RDF.type, y) in self.graph:
                        self.store_triple((u, RDF.type, xx))

        # RULE cls-avf
        elif p == OWL.allValuesFrom:
            xx, y = c, x
            for pp in self.graph.objects(xx, OWL.onProperty):
                for u in self.graph.subjects(RDF.type, xx):
                    for v in self.graph.objects(u, pp):
                        if self.restriction_typing_check(v, y):
                            self.store_triple((v, RDF.type, y))
                        else:
                            self.add_error(
                                "Violation of type restriction for allValuesFrom in %s for datatype %s on "
                                "value %s" % (pp, y, v)
                            )

        # RULES cls-hv1 and cls-hv2
        elif p == OWL.hasValue:
            xx, y = c, x
            for pp in self.graph.objects(xx, OWL.onProperty):
                # RULE cls-hv1
                for u in self.graph.subjects(RDF.type, xx):
                    self.store_triple((u, pp, y))
                # RULE cls-hv2
                for u in self.graph.subjects(pp, y):
                    self.store_triple((u, RDF.type, xx))

        # RULES cls-maxc1 and cls-maxc1
        elif p == OWL.maxCardinality:
            # This one is a bit complicated, because the literals have been
            # exchanged against bnodes...
            #
            # The construct should lead to an integer. Something may go wrong along the line
            # leading to an exception...
            xx = c
            if x.value == 0:
                # RULE cls-maxc1
                for pp in self.graph.objects(xx, OWL.onProperty):
                    for u, y in self.graph.subject_objects(pp):
                        # This should not occur:
                        if (u, RDF.type, xx) in self.graph:
                            self.add_error(
                                "Erroneous usage of maximum cardinality with %s and %s"
                                % (xx, y)
                            )
            elif x.value == 1:
                # RULE cls-maxc2
                for pp in self.graph.objects(xx, OWL.onProperty):
                    for u, y1 in self.graph.subject_objects(pp):
                        if (u, RDF.type, xx) in self.graph:
                            for y2 in self.graph.objects(u, pp):
                                if y1 != y2:
                                    self.store_triple((y1, OWL.sameAs, y2))

        # RULES cls-maxqc1, cls-maxqc2, cls-maxqc3, cls-maxqc4
        elif p == OWL.maxQualifiedCardinality:
            # This one is a bit complicated, because the literals have been
            # exchanged against bnodes...
            #
            # The construct should lead to an integer. Something may go wrong along the line
            # leading to an exception...
            xx = c
            if x.value == 0:
                # RULES cls-maxqc1 and cls-maxqc2 folded in one
                for pp in self.graph.objects(xx, OWL.onProperty):
                    for cc in self.graph.objects(xx, OWL.onClass):
                        for u, y in self.graph.subject_objects(pp):
                            # This should not occur:
                            if (
                                (y, RDF.type, cc) in self.graph or cc == OWL.Thing
                            ) and (u, RDF.type, xx) in self.graph:
                                self.add_error(
                                    "Erroneous usage of maximum qualified cardinality with %s, %s and %s"
                                    % (xx, cc, y)
                                )
            elif x.value == 1:
                # RULE cls-maxqc3 and cls-maxqc4 folded in one
                for pp in self.graph.objects(xx, OWL.onProperty):
                    for cc in self.graph.objects(xx, OWL.onClass):
                        for u, y1 in self.graph.subject_objects(pp):
                            if (u, RDF.type, xx) in self.graph:
                                if cc == OWL.Thing:
                                    for y2 in self.graph.objects(u, pp):
                                        if y1 != y2:
                                            self.store_triple((y1, OWL.sameAs, y2))
                                else:
                                    if (y1, RDF.type, cc) in self.graph:
                                        for y2 in self.graph.objects(u, pp):
                                            if (
                                                y1 != y2
                                                and (y2, RDF.type, cc) in self.graph
                                            ):
                                                self.store_triple((y1, OWL.sameAs, y2))

            # TODO: what if x.value not in (0, 1)? according to the spec
            # the cardinality shall be no more than 1, so add an # error?

        # RULE cls-oo
        elif p == OWL.oneOf:
            for y in self._list(x):
                self.store_triple((y, RDF.type, c))

    def _class_axioms(self, triple, cycle_num):
        """
        Table 7: Class Axioms. Essentially, the cax-* rules.
        @param triple: triple to work on
        @param cycle_num: which cycle are we in, starting with 1. Can be used for some optimization.
        """
        # In many of the 'if' branches, corresponding to rules in the document,
        # the branch begins by a renaming of variables (eg, pp, c = s, o).
        # There is no programming reasons for doing that, but by renaming the
        # variables it becomes easier to compare the declarative rules
        # in the document with the implementation
        c1, p, c2 = triple
        # RULE cax-sco
        if p == RDFS.subClassOf:
            # Other axioms sets classes to be subclasses of themselves, to one can optimize the trivial case
            if c1 != c2:
                for x in self.graph.subjects(RDF.type, c1):
                    self.store_triple((x, RDF.type, c2))

        # RULES cax-eqc1 and cax-eqc1
        # Other axioms set classes to be equivalent to themselves, one can optimize the trivial case
        elif p == OWL.equivalentClass and c1 != c2:
            # RULE cax-eqc1
            for x in self.graph.subjects(RDF.type, c1):
                self.store_triple((x, RDF.type, c2))
            # RULE cax-eqc1
            for x in self.graph.subjects(RDF.type, c2):
                self.store_triple((x, RDF.type, c1))

        # RULE cax-dw
        elif p == OWL.disjointWith:
            for x in self.graph.subjects(RDF.type, c1):
                if (x, RDF.type, c2) in self.graph:
                    self.add_error(
                        "Disjoint classes %s and %s have a common individual %s"
                        % (c1, c2, x)
                    )

        # RULE cax-adc
        elif p == RDF.type and c2 == OWL.AllDisjointClasses:
            x = c1
            for y in self.graph.objects(x, OWL.members):
                classes = self._list(y)
                if len(classes) > 0:
                    for i in range(0, len(classes) - 1):
                        cl1 = classes[i]
                        for z in self.graph.subjects(RDF.type, cl1):
                            for cl2 in classes[(i + 1) :]:
                                if (z, RDF.type, cl2) in self.graph:
                                    self.add_error(
                                        "Disjoint classes %s and %s have a common individual %s"
                                        % (cl1, cl2, z)
                                    )

    def _schema_vocabulary(self, triple, cycle_num):
        """
        Table 9: The Semantics of Schema Vocabulary. Essentially, the scm-* rules
        @param triple: triple to work on
        @param cycle_num: which cycle are we in, starting with 1. Can be used for some optimization.
        """
        # In many of the 'if' branches, corresponding to rules in the document,
        # the branch begins by a renaming of variables (eg, pp, c = s, o).
        # There is no programming reasons for doing that, but by renaming the
        # variables it becomes easier to compare the declarative rules
        # in the document with the implementation
        s, p, o = triple

        # RULE scm-cls
        if p == RDF.type and o == OWL.Class:
            c = s
            self.store_triple((c, RDFS.subClassOf, c))
            self.store_triple((c, OWL.equivalentClass, c))
            self.store_triple((c, RDFS.subClassOf, OWL.Thing))
            self.store_triple((OWL.Nothing, RDFS.subClassOf, c))

        # RULE scm-sco
        # Rule scm-eqc2
        elif p == RDFS.subClassOf:
            c1, c2 = s, o
            # RULE scm-sco
            # Optimize out the trivial identity case (set elsewhere already)
            if c1 != c2:
                for c3 in self.graph.objects(c2, RDFS.subClassOf):
                    # Another axiom already sets that...
                    if c1 != c3:
                        self.store_triple((c1, RDFS.subClassOf, c3))
            # RULE scm-eqc2
            if (c2, RDFS.subClassOf, c1) in self.graph:
                self.store_triple((c1, OWL.equivalentClass, c2))

        # RULE scm-eqc
        elif p == OWL.equivalentClass and s != o:
            c1, c2 = s, o
            self.store_triple((c1, RDFS.subClassOf, c2))
            self.store_triple((c2, RDFS.subClassOf, c1))

        # RULE scm-op and RULE scm-dp folded together
        # There is a bit of a cheating here: 'Property' is not, strictly speaking, in the rule set!
        elif p == RDF.type and (
            o == OWL.ObjectProperty or o == OWL.DatatypeProperty or o == RDF.Property
        ):
            pp = s
            self.store_triple((pp, RDFS.subPropertyOf, pp))
            self.store_triple((pp, OWL.equivalentProperty, pp))

        # RULE scm-spo
        # RULE scm-eqp2
        elif p == RDFS.subPropertyOf and s != o:
            p1, p2 = s, o
            # Optimize out the trivial identity case (set elsewhere already)
            # RULE scm-spo
            if p1 != p2:
                for p3 in self.graph.objects(p2, RDFS.subPropertyOf):
                    if p1 != p3:
                        self.store_triple((p1, RDFS.subPropertyOf, p3))

            # RULE scm-eqp2
            if (p2, RDFS.subPropertyOf, p1) in self.graph:
                self.store_triple((p1, OWL.equivalentProperty, p2))

        # RULE scm-eqp
        # Optimize out the trivial identity case (set elsewhere already)
        elif p == OWL.equivalentProperty and s != o:
            p1, p2 = s, o
            self.store_triple((p1, RDFS.subPropertyOf, p2))
            self.store_triple((p2, RDFS.subPropertyOf, p1))

        # RULES scm-dom1 and scm-dom2
        elif p == RDFS.domain:
            # RULE scm-dom1
            pp, c1 = s, o
            for (_x, _y, c2) in self.graph.triples((c1, RDFS.subClassOf, None)):
                if c1 != c2:
                    self.store_triple((pp, RDFS.domain, c2))
            # RULE scm-dom1
            p2, c = s, o
            for (p1, _x, _y) in self.graph.triples((None, RDFS.subPropertyOf, p2)):
                if p1 != p2:
                    self.store_triple((p1, RDFS.domain, c))

        # RULES scm-rng1 and scm-rng2
        elif p == RDFS.range:
            # RULE scm-rng1
            pp, c1 = s, o
            for (_x, _y, c2) in self.graph.triples((c1, RDFS.subClassOf, None)):
                if c1 != c2:
                    self.store_triple((pp, RDFS.range, c2))
            # RULE scm-rng1
            p2, c = s, o
            for (p1, _x, _y) in self.graph.triples((None, RDFS.subPropertyOf, p2)):
                if p1 != p2:
                    self.store_triple((p1, RDFS.range, c))

        # RULE scm-hv
        elif p == OWL.hasValue:
            c1, i = s, o
            for p1 in self.graph.objects(c1, OWL.onProperty):
                for c2 in self.graph.subjects(OWL.hasValue, i):
                    for p2 in self.graph.objects(c2, OWL.onProperty):
                        if (p1, RDFS.subPropertyOf, p2) in self.graph:
                            self.store_triple((c1, RDFS.subClassOf, c2))

        # RULES scm-svf1 and scm-svf2
        elif p == OWL.someValuesFrom:
            # RULE scm-svf1
            c1, y1 = s, o
            for pp in self.graph.objects(c1, OWL.onProperty):
                for c2 in self.graph.subjects(OWL.onProperty, pp):
                    for y2 in self.graph.objects(c2, OWL.someValuesFrom):
                        if (y1, RDFS.subClassOf, y2) in self.graph:
                            self.store_triple((c1, RDFS.subClassOf, c2))

            # RULE scm-svf2
            c1, y = s, o
            for p1 in self.graph.objects(c1, OWL.onProperty):
                for c2 in self.graph.subjects(OWL.someValuesFrom, y):
                    for p2 in self.graph.objects(c2, OWL.onProperty):
                        if (p1, RDFS.subPropertyOf, p2) in self.graph:
                            self.store_triple((c1, RDFS.subClassOf, c2))

        # RULES scm-avf1 and scm-avf2
        elif p == OWL.allValuesFrom:
            # RULE scm-avf1
            c1, y1 = s, o
            for pp in self.graph.objects(c1, OWL.onProperty):
                for c2 in self.graph.subjects(OWL.onProperty, pp):
                    for y2 in self.graph.objects(c2, OWL.allValuesFrom):
                        if (y1, RDFS.subClassOf, y2) in self.graph:
                            self.store_triple((c1, RDFS.subClassOf, c2))

            # RULE scm-avf2
            c1, y = s, o
            for p1 in self.graph.objects(c1, OWL.onProperty):
                for c2 in self.graph.subjects(OWL.allValuesFrom, y):
                    for p2 in self.graph.objects(c2, OWL.onProperty):
                        if (p1, RDFS.subPropertyOf, p2) in self.graph:
                            self.store_triple((c2, RDFS.subClassOf, c1))

        # RULE scm-int
        elif p == OWL.intersectionOf:
            c, x = s, o
            for ci in self._list(x):
                self.store_triple((c, RDFS.subClassOf, ci))

        # RULE scm-uni
        elif p == OWL.unionOf:
            c, x = s, o
            for ci in self._list(x):
                self.store_triple((ci, RDFS.subClassOf, c))



# -*- coding: utf-8 -*-
#
"""
This module is brute force implementation of the RDFS semantics on the top of RDFLib (with some caveats, see in the
introductory text).


**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""

__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from typing import Union

import rdflib
from rdflib import Literal, Graph
from rdflib.namespace import RDF, RDFS
from itertools import product
from owlrl.Closure import Core
from owlrl.AxiomaticTriples import RDFS_Axiomatic_Triples, RDFS_D_Axiomatic_Triples


######################################################################################################

# RDFS Semantics class
# noinspection PyPep8Naming
class RDFS_Semantics(Core):
    """
    RDFS Semantics class, ie, implementation of the RDFS closure graph.

    .. note:: Note that the module does *not* implement the so called Datatype entailment rules, simply because the
        underlying RDFLib does not implement the datatypes (ie, RDFLib will not make the literal "1.00" and "1.00000"
        identical, although even with all the ambiguities on datatypes, this I{should} be made equal...).

        Also, the so-called extensional entailment rules (Section 7.3.1 in the RDF Semantics document) have not been
        implemented either.

    The comments and references to the various rule follow the names as used in the `RDF Semantics document`_.

    .. _RDF Semantics document: http://www.w3.org/TR/rdf-mt/

    :param graph: The RDF graph to be extended.
    :type graph: :class:`rdflib.Graph`

    :param axioms: Whether (non-datatype) axiomatic triples should be added or not.
    :type axioms: bool

    :param daxioms: Whether datatype axiomatic triples should be added or not.
    :type daxioms: bool

    :param rdfs: Whether RDFS inference is also done (used in subclassed only).
    :type rdfs: bool
    """

    def __init__(self, graph: Graph, axioms, daxioms, rdfs: bool = False, destination: Union[None, Graph] = None):
        """
        @param graph: the RDF graph to be extended
        @type graph: rdflib.Graph
        @param axioms: whether (non-datatype) axiomatic triples should be added or not
        @type axioms: bool
        @param daxioms: whether datatype axiomatic triples should be added or not
        @type daxioms: bool
        @param rdfs: whether RDFS inference is also done (used in subclassed only)
        @type rdfs: boolean
        @param destination: the destination graph to which the results are written. If None, use the source graph.
        @type destination: rdflib.Graph
        """
        Core.__init__(self, graph, axioms, daxioms, rdfs=rdfs, destination=destination)

    def add_axioms(self):
        """
        Add axioms
        """
        for t in RDFS_Axiomatic_Triples:
            self.destination.add(t)
        for i in range(1, self.IMaxNum + 1):
            ci = RDF[("_%d" % i)]
            self.destination.add((ci, RDF.type, RDF.Property))
            self.destination.add((ci, RDFS.domain, RDFS.Resource))
            self.destination.add((ci, RDFS.range, RDFS.Resource))
            self.destination.add((ci, RDF.type, RDFS.ContainerMembershipProperty))

    def add_d_axioms(self):
        """
        This is not really complete, because it just uses the comparison possibilities that RDFLib provides.
        """
        # #1
        literals = (lt for lt in self._literals() if lt.datatype is not None)
        for lt in literals:
            self.destination.add((lt, RDF.type, lt.datatype))

        for t in RDFS_D_Axiomatic_Triples:
            self.destination.add(t)

    @staticmethod
    def _literals_same_as(lt1, lt2):
        if lt1.value is not None and lt2.value is not None:
            return lt1.value == lt2.value
        elif lt1.datatype is not None and lt2.datatype is not None:
            return lt1.__eq__(lt2)
        return False

    # noinspection PyBroadException
    def one_time_rules(self):
        """
        Some of the rules in the rule set are axiomatic in nature, meaning that they really have to be added only
        once, there is no reason to add these in a cycle. These are performed by this method that is invoked only once
        at the beginning of the process.

        In this case this is related to a 'hidden' same as rules on literals with identical values (though different
        lexical values).
        """
        # There is also a hidden sameAs rule in RDF Semantics: if a literal appears in a triple, and another one has
        # the same value, then the triple should be duplicated with the other value.
        literals = self._literals()
        items = (
            (lt1, lt2)
            for lt1, lt2 in product(literals, literals)
            if (lt1 is lt2) or self._literals_same_as(lt1, lt2)
        )
        for lt1, lt2 in items:
            # In OWL, this line is simply stating a sameAs for the
            # corresponding literals, and then let the usual rules take
            # effect. In RDFS this is not possible, so the sameAs rule is,
            # essentially replicated...
            for (s, p, o) in self.graph.triples((None, None, lt1)):
                self.destination.add((s, p, lt2))

    def rules(self, t, cycle_num):
        """
        Go through the RDFS entailment rules rdf1, rdfs4-rdfs12, by extending the graph.

        :param t: A triple (in the form of a tuple).
        :type t: tuple

        :param cycle_num: Which cycle are we in, starting with 1. Can be used for some (though minor) optimization.
        :type cycle_num: int
        """
        s, p, o = t
        # rdf1
        self.store_triple((p, RDF.type, RDF.Property))
        # rdfs4a
        if cycle_num == 1:
            self.store_triple((s, RDF.type, RDFS.Resource))
        # rdfs4b
        if cycle_num == 1:
            self.store_triple((o, RDF.type, RDFS.Resource))
        if p == RDFS.domain:
            # rdfs2
            for uuu, Y, yyy in self.graph.triples((None, s, None)):
                self.store_triple((uuu, RDF.type, o))
        if p == RDFS.range:
            # rdfs3
            for uuu, Y, vvv in self.graph.triples((None, s, None)):
                self.store_triple((vvv, RDF.type, o))
        if p == RDFS.subPropertyOf:
            # rdfs5
            for Z, Y, xxx in self.graph.triples((o, RDFS.subPropertyOf, None)):
                self.store_triple((s, RDFS.subPropertyOf, xxx))
            # rdfs7
            for zzz, Z, www in self.graph.triples((None, s, None)):
                self.store_triple((zzz, o, www))
        if p == RDF.type and o == RDF.Property:
            # rdfs6
            self.store_triple((s, RDFS.subPropertyOf, s))
        if p == RDF.type and o == RDFS.Class:
            # rdfs8
            self.store_triple((s, RDFS.subClassOf, RDFS.Resource))
            # rdfs10
            self.store_triple((s, RDFS.subClassOf, s))
        if p == RDFS.subClassOf:
            # rdfs9
            for vvv, Y, Z in self.graph.triples((None, RDF.type, s)):
                self.store_triple((vvv, RDF.type, o))
            # rdfs11
            for Z, Y, xxx in self.graph.triples((o, RDFS.subClassOf, None)):
                self.store_triple((s, RDFS.subClassOf, xxx))
        if p == RDF.type and o == RDFS.ContainerMembershipProperty:
            # rdfs12
            self.store_triple((s, RDFS.subPropertyOf, RDFS.member))
        if p == RDF.type and o == RDFS.Datatype:
            self.store_triple((s, RDFS.subClassOf, RDFS.Literal))

    def _literals(self):
        """
        Get all literals defined in the graph.
        """
        return set(
            o for s, p, o in self.graph.triples((None, None, None))
            if isinstance(o, Literal)
        )
